import matplotlib.pyplot as plt
import numpy as np
import os
from typing import Union, List, Dict
from sklearn.linear_model import Ridge
from kgcnn.data.utils import save_json_file, load_json_file
from kgcnn.data.transform.scaler.serial import deserialize


class ExtensiveMolecularScalerBase:
    """Scaler for extensive properties like energy to remove a simple linear behaviour with additive atom
    contributions.

    .. code-block:: python

        import numpy as np
        from kgcnn.scaler.mol import ExtensiveMolecularScalerBase
        data = np.random.rand(5).reshape((5,1))
        mol_num = [np.array([6, 1, 1, 1, 1]), np.array([7, 1, 1, 1]),
            np.array([6, 6, 1, 1, 1, 1]), np.array([6, 6, 1, 1]), np.array([6, 6, 1, 1, 1, 1, 1, 1])
        ]
        scaler = ExtensiveMolecularScalerBase()
        scaler.fit(data, atomic_number=mol_num)
        print(scaler.get_weights())
        print(scaler.get_config())
        scaler._plot_predict(data, mol_num)  # For debugging.

    """

    _attributes_list_sklearn = ["n_features_in_", "coef_", "intercept_", "n_iter_", "feature_names_in_"]
    _attributes_list_mol = ["scale_", "_fit_atom_selection", "_fit_atom_selection_mask"]
    max_atomic_number = 95

    def __init__(self, alpha: float = 1e-9, fit_intercept: bool = False, standardize_scale: bool = True, **kwargs):
        r"""Initialize scaler with parameters directly passed to scikit-learns :obj:`Ridge()`.

        Args:
            alpha (float): Regularization parameter for regression.
            fit_intercept (bool): Whether to allow a constant offset per target.
            standardize_scale (bool): Whether to standardize output after offset removal.
            kwargs: Additional arguments passed to :obj:`Ridge()`.
        """

        self.ridge = Ridge(alpha=alpha, fit_intercept=fit_intercept, **kwargs)
        self._standardize_scale = standardize_scale
        self._fit_atom_selection_mask = None
        self._fit_atom_selection = None
        self.scale_ = None

    def fit(self, molecular_property, atomic_number, sample_weight=None):
        r"""Fit atomic number to the molecular properties.

        Args:
            molecular_property (np.ndarray): Molecular properties of shape `(n_samples, n_properties)` .
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
            sample_weight: Sample weights `(n_samples,)` directly passed to :obj:`Ridge()`. Default is None.

        Returns:
            self
        """
        if len(atomic_number) != len(molecular_property):
            raise ValueError(
                "`ExtensiveMolecularScaler` different input shape '{0}' vs. '{1}'.".format(
                    len(atomic_number), len(molecular_property))
            )

        unique_number = [np.unique(x, return_counts=True) for x in atomic_number]
        all_unique = np.unique(np.concatenate([x[0] for x in unique_number], axis=0))
        self._fit_atom_selection = all_unique
        atom_mask = np.zeros(self.max_atomic_number, dtype="bool")
        atom_mask[all_unique] = True
        self._fit_atom_selection_mask = atom_mask
        total_number = []
        for unique_per_mol, num_unique in unique_number:
            array_atoms = np.zeros(self.max_atomic_number)
            array_atoms[unique_per_mol] = num_unique
            positives = array_atoms[atom_mask]
            total_number.append(positives)
        total_number = np.array(total_number)
        self.ridge.fit(total_number, molecular_property, sample_weight=sample_weight)
        diff = molecular_property - self.ridge.predict(total_number)
        if self._standardize_scale:
            self.scale_ = np.std(diff, axis=0)
        else:
            self.scale_ = np.ones(diff.shape[1:], dtype="float")
        return self

    def predict(self, atomic_number):
        """Predict the offset form atomic numbers. Requires :obj:`fit()` called previously.

        Args:
            atomic_number (list): List of array of atomic numbers. Example [np.array([7,1,1,1]), ...].

        Returns:
            np.ndarray: Offset of atomic properties fitted previously. Shape is `(n_samples, n_properties)`.
        """
        if self._fit_atom_selection_mask is None:
            raise ValueError("`ExtensiveMolecularScaler` has not been fitted yet. Can not predict.")
        unique_number = [np.unique(x, return_counts=True) for x in atomic_number]
        total_number = []
        for unique_per_mol, num_unique in unique_number:
            array_atoms = np.zeros(self.max_atomic_number)
            array_atoms[unique_per_mol] = num_unique
            positives = array_atoms[self._fit_atom_selection_mask]
            if np.sum(positives) != np.sum(num_unique):
                print("`ExtensiveMolecularScaler` got unknown atom species in transform.")
            total_number.append(positives)
        total_number = np.array(total_number)
        offset = self.ridge.predict(total_number)
        return offset

    def _plot_predict(self, molecular_property: np.ndarray, atomic_number: List[np.ndarray]):
        """Debug function to check prediction."""
        molecular_property = np.array(molecular_property)
        if len(molecular_property.shape) <= 1:
            molecular_property = np.expand_dims(molecular_property, axis=-1)
        predict_prop = self.predict(atomic_number)
        if len(predict_prop.shape) <= 1:
            predict_prop = np.expand_dims(predict_prop, axis=-1)
        mae = np.mean(np.abs(molecular_property - predict_prop), axis=0)
        fig = plt.figure()
        for i in range(predict_prop.shape[-1]):
            plt.scatter(predict_prop[:, i], molecular_property[:, i], alpha=0.3,
                        label="Pos: " + str(i) + " MAE: {0:0.4f} ".format(mae[i]))
        plt.plot(np.arange(np.amin(molecular_property), np.amax(molecular_property), 0.05),
                 np.arange(np.amin(molecular_property), np.amax(molecular_property), 0.05), color='red')
        plt.xlabel('Fitted')
        plt.ylabel('Actual')
        plt.legend(loc='upper left', fontsize='x-small')
        plt.show()
        return fig

    def transform(self, molecular_property, atomic_number, copy=True):
        """Transform any atomic number list with matching properties based on previous fit with sequential std-scaling.

        Args:
            molecular_property (np.ndarray): Molecular properties of shape `(n_samples, n_properties)` .
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
            copy (bool): Whether to copy or change in place.

        Returns:
            np.ndarray: Transformed atomic properties fitted. Shape is `(n_samples, n_properties)`.
        """
        if copy:
            molecular_property = molecular_property - self.predict(atomic_number)
            if self._standardize_scale:
                molecular_property = molecular_property / np.expand_dims(self.scale_, axis=0)
        else:
            molecular_property -= self.predict(atomic_number)
            if self._standardize_scale:
                molecular_property /= np.expand_dims(self.scale_, axis=0)
        return molecular_property

    def fit_transform(self, molecular_property, atomic_number, copy=True, sample_weight=None):
        """Combine fit and transform methods in one call.

        Args:
            molecular_property (np.ndarray): Molecular properties of shape `(n_samples, n_properties)` .
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
            copy (bool): Whether to copy or change in place.
            sample_weight: Sample weights `(n_samples,)` directly passed to :obj:`Ridge()`. Default is None.

        Returns:
            np.ndarray: Transformed atomic properties fitted. Shape is `(n_samples, n_properties)`.
        """
        self.fit(molecular_property=molecular_property, atomic_number=atomic_number, sample_weight=sample_weight)
        return self.transform(molecular_property=molecular_property, atomic_number=atomic_number, copy=copy)

    def inverse_transform(self, molecular_property: np.ndarray, atomic_number: List[np.ndarray],
                          copy: bool = True) -> np.ndarray:
        """Reverse the transform method to original properties without offset removed and scaled to original units.

        Args:
            molecular_property (np.ndarray): Molecular properties of shape `(n_samples, n_properties)` .
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
            copy (bool): Whether to copy or change in place.

        Returns:
            np.ndarray: Original atomic properties. Shape is `(n_samples, n_properties)`.
        """
        if copy:
            if self._standardize_scale:
                molecular_property = molecular_property * np.expand_dims(self.scale_, axis=0)
            molecular_property = molecular_property + self.predict(atomic_number)
        else:
            if self._standardize_scale:
                molecular_property *= np.expand_dims(self.scale_, axis=0)
            molecular_property += self.predict(atomic_number)
        return molecular_property

    def get_config(self):
        """Get configuration for scaler."""
        config = {}
        config.update(self.ridge.get_params())
        config.update({"standardize_scale": self._standardize_scale})
        return config

    def set_config(self, config):
        """Set configuration for scaler.

        Args:
            config (dict): Config dictionary.
        """
        self._standardize_scale = config["standardize_scale"]
        config_ridge = {key: value for key, value in config.items() if key not in ["standardize_scale"]}
        self.ridge.set_params(**config_ridge)
        return self

    def get_weights(self) -> dict:
        """Get weights for this scaler after fit."""
        weights = dict()
        for x in self._attributes_list_mol:
            weights.update({x: np.array(getattr(self, x)).tolist()})
        for x in self._attributes_list_sklearn:
            if hasattr(self.ridge, x):
                weights.update({x: np.array(getattr(self.ridge, x)).tolist()})
        return weights

    def set_weights(self, weights: dict):
        """Set weights for this scaler.

        Args:
            weights (dict): Weight dictionary.
        """
        for item, value in weights.items():
            if item in self._attributes_list_mol:
                setattr(self, item, np.array(value))
            elif item in self._attributes_list_sklearn:
                setattr(self.ridge, item, np.array(value))
            else:
                print("`ExtensiveMolecularScaler` got unknown weight '%s'." % item)

    def save_weights(self, file_path: str):
        """Save weights as numpy to file.

        Args:
            file_path: Filepath to save weights.
        """
        weights = self.get_weights()
        # Make them all numpy arrays for save.
        for key, value in weights.items():
            weights[key] = np.array(value)
        if len(weights) > 0:
            np.savez(os.path.splitext(file_path)[0] + ".npz", **weights)
        else:
            print("Error no weights to save for `ExtensiveMolecularScaler`.")

    def get_scaling(self):
        """Get scale of shape (1, n_properties)."""
        if self.scale_ is None:
            return
        return np.expand_dims(self.scale_, axis=0)

    def save(self, file_path: str):
        """Save scaler serialization to file.

        Args:
            file_path: Filepath to save scaler serialization.
        """
        conf = self.get_config()
        weights = self.get_weights()
        full_info = {"class_name": type(self).__name__, "module_name": type(self).__module__,
                     "config": conf, "weights": weights}
        save_json_file(full_info, os.path.splitext(file_path)[0] + ".json")

    def load(self, file_path: str):
        """Load scaler serialization from file.

        Args:
            file_path: Filepath to load scaler serialization.
        """
        full_info = load_json_file(file_path)
        # Could verify class_name and module_name here.
        self.set_config(full_info["config"])
        self.set_weights(full_info["weights"])
        return self

    # Similar functions that work on dataset plus property names.
    # noinspection PyPep8Naming
    def fit_dataset(self, dataset: List[Dict[str, np.ndarray]],
                    molecular_property: str, atomic_number: str, sample_weight: str = None):
        r"""Fit to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            molecular_property (str): Name of molecular property information in dataset. For example "energy".
            atomic_number (str): Name of atomic number information in dataset. For example "atomic number".
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".

        Returns:
            self.
        """
        return self.fit(
            molecular_property=np.array([item[molecular_property] for item in dataset]),
            atomic_number=[item[atomic_number] for item in dataset],
            sample_weight=[item[sample_weight] for item in dataset] if sample_weight is not None else None
        )

    # noinspection PyPep8Naming
    def transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                          molecular_property: str, atomic_number: str, copy: bool = True,
                          copy_dataset: bool = False,
                          ) -> List[Dict[str, np.ndarray]]:
        r"""Transform dataset with relevant `X` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            molecular_property (str): Name of molecular property information in dataset. For example "energy".
            atomic_number (str): Name of atomic number information in dataset. For example "atomic number".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        if copy_dataset:
            dataset = dataset.copy()
        out = self.transform(
            molecular_property=np.array([item[molecular_property] for item in dataset]),
            atomic_number=[item[atomic_number] for item in dataset],
            copy=copy,
        )
        for graph, out_value in zip(dataset, out):
            graph[molecular_property] = out_value
        return dataset

    # noinspection PyPep8Naming
    def inverse_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                                  molecular_property: str, atomic_number: str, copy: bool = True,
                                  copy_dataset: bool = False,
                                  ) -> List[Dict[str, np.ndarray]]:
        r"""Inverse transform dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            molecular_property (str): Name of molecular property information in dataset. For example "energy".
            atomic_number (str): Name of atomic number information in dataset. For example "atomic number".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Inverse-transformed dataset.
        """
        if copy_dataset:
            dataset = dataset.copy()
        out = self.inverse_transform(
            molecular_property=np.array([item[molecular_property] for item in dataset]),
            atomic_number=[item[atomic_number] for item in dataset],
            copy=copy,
        )
        for graph, out_value in zip(dataset, out):
            graph[molecular_property] = out_value
        return dataset

    # noinspection PyPep8Naming
    def fit_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                              molecular_property: str, atomic_number: str,
                              sample_weight: str = None, copy: bool = True, copy_dataset: bool = False
                              ) -> List[Dict[str, np.ndarray]]:
        r"""Fit and transform to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            molecular_property (str): Name of molecular property information in dataset. For example "energy".
            atomic_number (str): Name of atomic number information in dataset. For example "atomic number".
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        self.fit_dataset(dataset=dataset, molecular_property=molecular_property, atomic_number=atomic_number,
                         sample_weight=sample_weight)
        return self.transform_dataset(dataset=dataset, molecular_property=molecular_property,
                                      atomic_number=atomic_number, copy=copy, copy_dataset=copy_dataset)


class ExtensiveMolecularScaler(ExtensiveMolecularScalerBase):
    r"""Scaler for extensive properties like energy to remove a simple linear behaviour with additive atom
    contributions. Interface is designed after scikit-learn scaler. Internally Ridge regression ist used.
    Only the atomic number is used as extensive scaler. This could be further improved by also taking bonds and
    interactions into account, e.g. as energy contribution.

    .. code-block:: python

        import numpy as np
        from kgcnn.scaler.mol import ExtensiveMolecularScaler
        data = np.random.rand(5).reshape((5,1))
        mol_num = [np.array([6, 1, 1, 1, 1]), np.array([7, 1, 1, 1]),
            np.array([6, 6, 1, 1, 1, 1]), np.array([6, 6, 1, 1]), np.array([6, 6, 1, 1, 1, 1, 1, 1])
        ]
        scaler = ExtensiveMolecularScaler()
        scaler.fit(X=data, atomic_number=mol_num)
        print(scaler.get_weights())
        print(scaler.get_config())
        scaler._plot_predict(data, mol_num)  # For debugging.
        print(scaler.inverse_transform(scaler.transform(X=data, atomic_number=mol_num), atomic_number=mol_num))
        print(data)
        scaler.save("example.json")
        new_scaler = ExtensiveMolecularScaler()
        new_scaler.load("example.json")
        print(scaler.inverse_transform(scaler.transform(X=data, atomic_number=mol_num), atomic_number=mol_num))

    """

    # noinspection PyPep8Naming
    def fit(self, X, *, y: Union[None, np.ndarray] = None, sample_weight=None, atomic_number=None):
        r"""Fit atomic number to the molecular properties.

        Args:
            X (np.ndarray): Array of atomic properties of shape `(n_samples, n_properties)`.
            y (np.ndarray): Ignored.
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
            sample_weight: Sample weights `(n_samples,)` directly passed to :obj:`Ridge()`. Default is None.

        Returns:
            self.
        """
        return super(ExtensiveMolecularScaler, self).fit(
            molecular_property=X, atomic_number=atomic_number, sample_weight=sample_weight)

    # noinspection PyPep8Naming
    def transform(self, X, *, y=None, copy=True, atomic_number=None):
        """Transform any atomic number list with matching properties based on previous fit with sequential std-scaling.

        Args:
            X (np.ndarray): Array of atomic properties of shape `(n_samples, n_properties)`.
            y (np.ndarray): Ignored.
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
            copy (bool): Whether to copy or change in place.

        Returns:
            np.ndarray: Transformed atomic properties fitted. Shape is `(n_samples, n_properties)`.
        """
        return super(ExtensiveMolecularScaler, self).transform(
            molecular_property=X, atomic_number=atomic_number, copy=copy)

    # noinspection PyPep8Naming
    def fit_transform(self, X, *, y=None, copy=True, atomic_number=None, sample_weight=None):
        r"""Fit and transform.

        Args:
            X (np.ndarray): Array of atomic properties of shape `(n_samples, n_properties)`.
            y (np.ndarray): Ignored.
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
            copy (bool): Whether to copy or change in place.
            sample_weight: Sample weights `(n_samples,)` directly passed to :obj:`Ridge()`. Default is None.

        Returns:
            np.ndarray: Transformed properties.
        """
        self.fit(X=X, y=y, sample_weight=sample_weight, atomic_number=atomic_number)
        return self.transform(X=X, y=y, copy=copy, atomic_number=atomic_number)

    # noinspection PyPep8Naming
    def inverse_transform(self, X, *, y=None, copy=True, atomic_number=None):
        """Reverse the transform method to original properties without offset removed and scaled to original units.

        Args:
            X (np.ndarray): Array of atomic properties of shape `(n_samples, n_properties)`.
            y (np.ndarray): Ignored.
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
            copy (bool): Whether to copy or change in place.

        Returns:
            np.ndarray: Original atomic properties. Shape is `(n_samples, n_properties)`.
        """

        return super(ExtensiveMolecularScaler, self).inverse_transform(
            molecular_property=X, atomic_number=atomic_number, copy=copy)

    # Similar functions that work on dataset plus property names.
    # noinspection PyPep8Naming
    def fit_dataset(self, dataset: List[Dict[str, np.ndarray]], X: str = None, *, y: str = None,
                    atomic_number: str = None, sample_weight: str = None):
        r"""Fit to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            X (str): Name of X information in dataset. For example "graph_properties".
            y (str): Not used.
            atomic_number (str): Name of atomic number information in dataset. For example "atomic number".
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".

        Returns:
            self.
        """
        return super(ExtensiveMolecularScaler, self).fit_dataset(
            dataset=dataset, molecular_property=X, atomic_number=atomic_number, sample_weight=sample_weight)

    # noinspection PyPep8Naming
    def transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                          X: str = None, *, atomic_number: str = None, copy: bool = True,
                          copy_dataset: bool = False,
                          ) -> List[Dict[str, np.ndarray]]:
        r"""Transform dataset with relevant `X` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            X (str): Name of X information in dataset. For example "graph_properties".
            atomic_number (str): Name of atomic number information in dataset. For example "atomic number".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        return super(ExtensiveMolecularScaler, self).transform_dataset(
            dataset=dataset, molecular_property=X, atomic_number=atomic_number, copy_dataset=copy_dataset, copy=copy)

    # noinspection PyPep8Naming
    def inverse_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                                  X: str = None, *, atomic_number: str = None, copy: bool = True,
                                  copy_dataset: bool = False,
                                  ) -> List[Dict[str, np.ndarray]]:
        r"""Inverse transform dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            X (str): Name of X information in dataset. For example "graph_properties".
            atomic_number (str): Name of atomic number information in dataset. For example "atomic number".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Inverse-transformed dataset.
        """
        return super(ExtensiveMolecularScaler, self).inverse_transform_dataset(
            dataset=dataset, molecular_property=X, atomic_number=atomic_number, copy_dataset=copy_dataset, copy=copy)

    # noinspection PyPep8Naming
    def fit_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                              X: str = None, *, y: str = None, atomic_number: str = None,
                              sample_weight: str = None, copy: bool = True, copy_dataset: bool = False
                              ) -> List[Dict[str, np.ndarray]]:
        r"""Fit and transform to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            X (str): Name of X information in dataset. For example "graph_properties".
            y (str): Not used.
            atomic_number (str): Name of atomic number information in dataset. For example "atomic number".
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        self.fit_dataset(dataset=dataset, X=X, y=y, atomic_number=atomic_number, sample_weight=sample_weight)
        return self.transform_dataset(dataset=dataset, X=X, atomic_number=atomic_number,
                                      copy=copy, copy_dataset=copy_dataset)


class ExtensiveMolecularLabelScaler(ExtensiveMolecularScalerBase):
    r"""Equivalent of :obj:`ExtensiveMolecularScaler` for labels, which uses the `y` argument for labels.
    For `X` the atomic numbers can be passed.

    .. code-block:: python

        import numpy as np
        from kgcnn.scaler.mol import ExtensiveMolecularLabelScaler
        data = np.random.rand(5).reshape((5,1))
        mol_num = [np.array([6, 1, 1, 1, 1]), np.array([7, 1, 1, 1]),
            np.array([6, 6, 1, 1, 1, 1]), np.array([6, 6, 1, 1]), np.array([6, 6, 1, 1, 1, 1, 1, 1])
        ]
        scaler = ExtensiveMolecularLabelScaler()
        scaler.fit(X=mol_num, y=data)
        print(scaler.get_weights())
        print(scaler.get_config())
        scaler._plot_predict(data, mol_num)  # For debugging.
        print(scaler.inverse_transform(X=mol_num, y=scaler.transform(X=mol_num, y=data)))
        print(data)
        scaler.save("example.json")
        new_scaler = ExtensiveMolecularLabelScaler()
        new_scaler.load("example.json")
        print(scaler.inverse_transform(X=mol_num, y=scaler.transform(X=mol_num, y=data)))

    """

    def _assert_has_y(self, y):
        if y is None:
            raise ValueError(
                "Require labels in `y` for `%s`. Input must be e.g. 'fit(y=data)'." % type(self).__name__)

    # noinspection PyPep8Naming
    def fit(self, y: Union[None, list, np.ndarray] = None, *, X=None, sample_weight=None, atomic_number=None):
        """Fit labels with atomic number information.

        Args:
            y: Array of atomic labels of shape `(n_samples, n_labels)`.
            X: List of array of atomic numbers. Example [np.array([7,1,1,1]), ...].
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
                Optional, since they should be contained in `X` . Note that if assigning `atomic_numbers`
                then `X` is ignored.
            sample_weight: Sample weights `(n_samples,)` directly passed to :obj:`Ridge()`. Default is None.

        Returns:
            np.ndarray: Transformed y.
        """
        self._assert_has_y(y)
        atomic_number = atomic_number if atomic_number else X
        return super(ExtensiveMolecularLabelScaler, self).fit(
            molecular_property=y, sample_weight=sample_weight, atomic_number=atomic_number)

    # noinspection PyPep8Naming
    def transform(self, y=None, *, X, copy=True, atomic_number=None):
        """Transform any atomic number list with matching labels based on previous fit with sequential std-scaling.

        Args:
            y: Array of atomic labels of shape `(n_samples, n_labels)`.
            X: List of array of atomic numbers. Example [np.array([7,1,1,1]), ...].
            copy (bool): Whether to copy or change in place.
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
                Optional, since they should be contained in `X` . Note that if assigning `atomic_numbers`
                then `X` is ignored.

        Returns:
            np.ndarray: Transformed y.
        """
        self._assert_has_y(y)
        atomic_number = atomic_number if atomic_number else X
        return super(ExtensiveMolecularLabelScaler, self).transform(
            molecular_property=y, atomic_number=atomic_number, copy=copy)

    # noinspection PyPep8Naming
    def fit_transform(self, y=None, *, X=None, copy=True, atomic_number=None, sample_weight=None):
        """Fit and transform.

        Args:
            y: Array of atomic labels of shape `(n_samples, n_labels)`.
            X: List of array of atomic numbers. Example [np.array([7,1,1,1]), ...].
            copy (bool): Whether to copy or change in place.
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
                Optional, since they should be contained in `X` . Note that if assigning `atomic_numbers`
                then `X` is ignored.
            sample_weight: Sample weights `(n_samples,)` directly passed to :obj:`Ridge()`. Default is None.

        Returns:

        """
        self.fit(y=y, X=X, sample_weight=sample_weight, atomic_number=atomic_number)
        return self.transform(y=y, X=X, copy=copy, atomic_number=atomic_number)

    # noinspection PyPep8Naming
    def inverse_transform(self, y=None, *, X=None, copy=True, atomic_number=None):
        """Reverse the transform method to original labels without offset removed and scaled to original units.

        Args:
            y: Array of atomic labels of shape `(n_samples, n_labels)`.
            X: List of array of atomic numbers. Example [np.array([7,1,1,1]), ...].
            copy (bool): Whether to copy or change in place.
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
                Optional, since they should be contained in `X` . Note that if assigning `atomic_numbers`
                then `X` is ignored.

        Returns:
            np.ndarray: Transformed y.
        """
        self._assert_has_y(y)
        atomic_number = atomic_number if atomic_number else X
        return super(ExtensiveMolecularLabelScaler, self).inverse_transform(
            molecular_property=y, atomic_number=atomic_number, copy=copy)

    # Similar functions that work on dataset plus property names.
    # noinspection PyPep8Naming
    def fit_dataset(self, dataset: List[Dict[str, np.ndarray]], y: str = None, *, X: str = None,
                    atomic_number: str = None, sample_weight: str = None):
        r"""Fit to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            y (str): Name of y information in dataset. For example "graph_labels".
            X (str): Not used. Optional atomic number information, if `atomic_number` is not set.
            atomic_number (str): Name of atomic number information in dataset. For example "atomic number".
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".

        Returns:
            self.
        """
        # Just changing order of x,y here.
        atomic_number = atomic_number if atomic_number else X
        return super(ExtensiveMolecularLabelScaler, self).fit_dataset(
            dataset=dataset, molecular_property=y, atomic_number=atomic_number, sample_weight=sample_weight)

    # noinspection PyPep8Naming
    def transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                          y: str = None, *, X: str = None, atomic_number: str = None, copy: bool = True,
                          copy_dataset: bool = False,
                          ) -> List[Dict[str, np.ndarray]]:
        r"""Transform dataset with relevant `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            y (str): Name of y information in dataset. For example "graph_labels".
            X (str): Not used. Optional atomic number information, if `atomic_number` is not set.
            atomic_number (str): Name of atomic number information in dataset. For example "atomic number".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        # Just changing order of x,y here.
        atomic_number = atomic_number if atomic_number else X
        return super(ExtensiveMolecularLabelScaler, self).transform_dataset(
            dataset=dataset, molecular_property=y, atomic_number=atomic_number, copy_dataset=copy_dataset, copy=copy)

    # noinspection PyPep8Naming
    def inverse_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                                  y: str = None, *, X: str = None,
                                  atomic_number: str = None, copy: bool = True,
                                  copy_dataset: bool = False,
                                  ) -> List[Dict[str, np.ndarray]]:
        r"""Inverse transform dataset with relevant `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            y (str): Name of y information in dataset. For example "graph_labels".
            X (str): Not used. Optional atomic number information, if `atomic_number` is not set.
            atomic_number (str): Name of atomic number information in dataset. For example "atomic number".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Inverse-transformed dataset.
        """
        atomic_number = atomic_number if atomic_number else X
        return super(ExtensiveMolecularLabelScaler, self).inverse_transform_dataset(
            dataset=dataset, molecular_property=y, atomic_number=atomic_number, copy_dataset=copy_dataset, copy=copy)

    # noinspection PyPep8Naming
    def fit_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                              y: str = None, *, X: str = None, atomic_number: str = None,
                              sample_weight: str = None, copy: bool = True, copy_dataset: bool = False
                              ) -> List[Dict[str, np.ndarray]]:
        r"""Fit and transform to dataset with relevant `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            y (str): Name of y information in dataset. For example "graph_properties".
            X (str): Not used. Optional atomic number information, if `atomic_number` is not set.
            atomic_number (str): Name of atomic number information in dataset. For example "atomic number".
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        # Uses the functions of this class.
        self.fit_dataset(dataset=dataset, y=y, X=X, atomic_number=atomic_number, sample_weight=sample_weight)
        return self.transform_dataset(
            dataset=dataset, y=y, X=X, atomic_number=atomic_number, copy=copy, copy_dataset=copy_dataset)


class QMGraphLabelScaler:
    r"""A scaler that scales QM targets differently. For now, the main difference is that intensive and extensive
    properties are scaled differently. In principle, also dipole, polarizability or rotational constants
    could to be standardized differently. Interface is designed after scikit-learn scaler.

    The class is simply a list of separate scaler and scales each target of shape [N_samples, target] with a scaler
    from its list. :obj:`QMGraphLabelScaler` is intended as a scaler list class.

    .. note::

        The scaler uses `y` argument

    Each label is passed to the corresponding scaler in list simply as first argument without keyword `X` or `y`.

    .. code-block:: python

        import numpy as np
        from kgcnn.scaler.mol import QMGraphLabelScaler, ExtensiveMolecularScaler
        from kgcnn.scaler.scaler import StandardScaler
        data = np.random.rand(10).reshape((5,2))
        mol_num = [np.array([6, 1, 1, 1, 1]), np.array([7, 1, 1, 1]),
            np.array([6, 6, 1, 1, 1, 1]), np.array([6, 6, 1, 1]), np.array([6, 6, 1, 1, 1, 1, 1, 1])
        ]
        scaler = QMGraphLabelScaler([ExtensiveMolecularScaler(), StandardScaler()])
        scaler.fit(y=data, atomic_number=mol_num)
        print(scaler.get_weights())
        print(scaler.get_config())
        print(scaler.inverse_transform(scaler.transform(y=data, atomic_number=mol_num), atomic_number=mol_num))
        print(data)
        scaler.save("example.json")
        new_scaler = QMGraphLabelScaler([ExtensiveMolecularScaler(), StandardScaler()])
        new_scaler.load("example.json")
        print(new_scaler.inverse_transform(scaler.transform(y=data, atomic_number=mol_num), atomic_number=mol_num))

    """

    def __init__(self, scaler: list):

        if not isinstance(scaler, list):
            raise TypeError("Scaler information for `QMGraphLabelScaler` must be list, got '%s'." % scaler)

        self.scaler_list = []
        for x in scaler:
            if hasattr(x, "transform") and hasattr(x, "fit"):
                self.scaler_list.append(x)
            elif isinstance(x, dict):
                # Otherwise, must be serialized version of a scaler.
                self.scaler_list.append(deserialize(x))
            else:
                raise ValueError("Unsupported scaler type '%s'." % x)

    # noinspection PyPep8Naming
    def fit_transform(self, y: Union[np.ndarray, List[np.ndarray]] = None,
                      X: Union[np.ndarray, List[np.ndarray], None] = None,
                      atomic_number: Union[np.ndarray, List[np.ndarray], None] = None,
                      copy: bool = True,
                      sample_weight=None):
        r"""Fit and transform all target labels for QM.

        Args:
            y (np.ndarray): Array of atomic labels of shape `(n_samples, n_labels)`.
            X (np.ndarray): Not used.
            copy (bool): Whether to copy or change in place.
            atomic_number (list): List of array of atomic numbers. Example [np.array([7,1,1,1]), ...].
            sample_weight: Sample weights `(n_samples,)` directly passed to :obj:`Ridge()`. Default is None.

        Returns:
            np.ndarray: Transformed labels of shape `(n_samples, n_labels)`.
        """
        self.fit(y=y, X=X, atomic_number=atomic_number, sample_weight=sample_weight)
        return self.transform(y=y, X=X, copy=copy, atomic_number=atomic_number)

    # noinspection PyPep8Naming
    def transform(self, y: Union[np.ndarray, List[np.ndarray]] = None,
                  X: Union[np.ndarray, List[np.ndarray], None] = None,
                  atomic_number: Union[np.ndarray, List[np.ndarray], None] = None,
                  copy=True):
        r"""Transform all target labels for QM. Requires :obj:`fit()` called previously.

        Args:
            y (np.ndarray): Array of QM unscaled labels of shape `(n_samples, n_labels)`.
            X (np.ndarray): Not used.
            copy (bool): Whether to copy or change in place.
            atomic_number (list): List of atomic numbers for each molecule. E.g. `[np.array([6,1,1,1]), ...]`.

        Returns:
            np.ndarray: Transformed labels of shape `(n_samples, n_labels)`.
        """
        labels, atomic_number = self._check_input(atomic_number, X, y)

        if copy:
            out_labels = []
            for i, x in enumerate(self.scaler_list):
                out_labels.append(x.transform(labels[:, i:i + 1], atomic_number=atomic_number, copy=copy))
            out_labels = np.concatenate(out_labels, axis=-1)
        else:
            for i, x in enumerate(self.scaler_list):
                x.transform(labels[:, i:i + 1], atomic_number=atomic_number, copy=copy)
            out_labels = labels
        return out_labels

    # noinspection PyPep8Naming
    def fit(self, y: Union[np.ndarray, List[np.ndarray]] = None,
            X: Union[np.ndarray, List[np.ndarray], None] = None,
            atomic_number: Union[np.ndarray, List[np.ndarray], None] = None,
            sample_weight=None):
        r"""Fit scaling of QM graph labels or targets.

        Args:
            y (np.ndarray): Array of atomic labels of shape `(n_samples, n_labels)`.
            X (np.ndarray): Ignored.
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].
            sample_weight: Sample weights `(n_samples,)` directly passed to :obj:`Ridge()`. Default is None.

        Returns:
            self
        """
        labels, atomic_number = self._check_input(atomic_number, X, y)

        for i, x in enumerate(self.scaler_list):
            x.fit(labels[:, i:i + 1], atomic_number=atomic_number, sample_weight=sample_weight)

        return self

    # noinspection PyPep8Naming
    def inverse_transform(self, y: Union[np.ndarray, List[np.ndarray]] = None,
                          X: Union[np.ndarray, List[np.ndarray], None] = None,
                          atomic_number: Union[np.ndarray, List[np.ndarray], None] = None,
                          copy: bool = True):
        r"""Back-transform all target labels for QM.

        Args:
            y (np.ndarray): Array of atomic labels of shape `(n_samples, n_labels)`.
            X (np.ndarray): Not used.
            copy (bool): Whether to copy or change in place.
            atomic_number (list): List of arrays of atomic numbers. Example [np.array([7,1,1,1]), ...].

        Returns:
            np.ndarray: Back-transformed labels of shape `(n_samples, n_labels)`.
        """
        labels, atomic_number = self._check_input(atomic_number, X, y)

        if copy:
            out_labels = []
            for i, x in enumerate(self.scaler_list):
                out_labels.append(x.inverse_transform(labels[:, i:i + 1], atomic_number=atomic_number, copy=copy))
            out_labels = np.concatenate(out_labels, axis=-1)
        else:
            for i, x in enumerate(self.scaler_list):
                x.inverse_transform(labels[:, i:i + 1], atomic_number=atomic_number, copy=copy)
            out_labels = labels

        return out_labels

    # noinspection PyPep8Naming
    def _check_input(self, node_number, X, y):
        assert X is not None or y is not None, "`QMGraphLabelScaler` did not get properties or labels."
        graph_labels = X if (y is None and node_number is not None) else y
        node_number = node_number if node_number is not None else X
        assert len(node_number) == len(graph_labels), "`QMGraphLabelScaler` input length does not match."
        assert graph_labels.shape[-1] == len(self.scaler_list), "`QMGraphLabelScaler` got wrong number of labels."
        return graph_labels, node_number

    @property
    def scale_(self):
        """Composite scale of all scaler in list."""
        return np.concatenate([x.scale_ for x in self.scaler_list], axis=0)

    def get_scaling(self):
        """Get scale of shape (1, n_properties)."""
        return np.expand_dims(self.scale_, axis=0)

    def get_weights(self):
        """Get weights for this scaler after fit."""
        weights = {"scaler": [x.get_weights() for x in self.scaler_list]}
        return weights

    def set_weights(self, weights: dict):
        """Set weights for this scaler.

        Args:
            weights (dict): Weight dictionary.
        """
        scaler_weights = weights["scaler"]
        for i, x in enumerate(scaler_weights):
            self.scaler_list[i].set_weights(x)

    def save_weights(self, file_path: str):
        """Save weights as numpy to file.

        Args:
            file_path: Filepath to save weights.
        """
        all_weights = {}
        for i, x in enumerate(self.scaler_list):
            w = x.get_weights()
            for key, value in w.items():
                all_weights[str(key) + "_%i" % i] = np.array(value)
        np.savez(os.path.splitext(file_path)[0] + ".npz", **all_weights)

    def get_config(self):
        """Get configuration for scaler."""
        config = {"scaler": [
            {"class_name": type(x).__name__, "module_name": type(x).__module__,
             "config": x.get_config()} for x in self.scaler_list]
        }
        return config

    def set_config(self, config):
        """Set configuration for scaler.

        Args:
            config (dict): Config dictionary.
        """
        scaler_conf = config["scaler"]
        for i, x in enumerate(scaler_conf):
            self.scaler_list[i].set_config(x["config"])
        return self

    def save(self, file_path: str):
        """Save scaler serialization to file.

        Args:
            file_path: Filepath to save scaler serialization.
        """
        conf = self.get_config()
        weights = self.get_weights()
        full_info = {"class_name": type(self).__name__, "module_name": type(self).__module__,
                     "config": conf, "weights": weights}
        save_json_file(full_info, os.path.splitext(file_path)[0] + ".json")

    def load(self, file_path: str):
        """Load scaler serialization from file.

        Args:
            file_path: Filepath to load scaler serialization.
        """
        full_info = load_json_file(file_path)
        # Could verify class_name and module_name here.
        self.set_config(full_info["config"])
        self.set_weights(full_info["weights"])
        return self

    # Similar functions that work on dataset plus property names.
    # noinspection PyPep8Naming
    def fit_dataset(self, dataset: List[Dict[str, np.ndarray]],
                    y: str, *,
                    X: str = None,
                    atomic_number: str = None,
                    sample_weight: str = None):
        r"""Fit to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            y (str): Name of label information in dataset. For example "graph_labels".
            X (str): Optional X information in dataset. For example "atomic_number".
            atomic_number (list): Name of atomic number information in dataset. For example "atomic_number".
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".

        Returns:
            self.
        """
        return self.fit(
            y=[item[y] for item in dataset],
            X=[item[X] for item in dataset] if X is not None else None,
            atomic_number=[item[atomic_number] for item in dataset] if atomic_number is not None else None,
            sample_weight=[item[sample_weight] for item in dataset] if sample_weight is not None else None
        )

    # noinspection PyPep8Naming
    def transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                          y: str, *,
                          X: str = None, atomic_number: str = None,
                          copy: bool = True,
                          copy_dataset: bool = False,
                          ) -> List[Dict[str, np.ndarray]]:
        r"""Transform dataset with relevant `X` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            y (str): Name of y information in dataset. For example "graph_properties".
            X (str): Optional X information in dataset. For example "atomic_number".
            atomic_number (list): Name of atomic number information in dataset. For example "atomic_number".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        if copy_dataset:
            dataset = dataset.copy()
        out = self.transform(
            y=[graph[y] for graph in dataset],
            X=[item[X] for item in dataset] if X is not None else None,
            atomic_number=[item[atomic_number] for item in dataset] if atomic_number is not None else None,
            copy=copy,
        )
        for graph, out_value in zip(dataset, out):
            graph[y] = out_value
        return dataset

    # noinspection PyPep8Naming
    def inverse_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                                  y: str = None, *,
                                  X: str = None,
                                  atomic_number: str = None,
                                  copy: bool = True,
                                  copy_dataset: bool = False,
                                  ) -> List[Dict[str, np.ndarray]]:
        r"""Inverse transform dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            y (str): Name of y information in dataset. For example "graph_properties".
            X (str): Optional X information in dataset. For example "atomic_number".
            atomic_number (list): Name of atomic number information in dataset. For example "atomic_number".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Inverse-transformed dataset.
        """
        if copy_dataset:
            dataset = dataset.copy()
        out = self.inverse_transform(
            y=[graph[y] for graph in dataset],
            X=[item[X] for item in dataset] if X is not None else None,
            atomic_number=[item[atomic_number] for item in dataset] if atomic_number is not None else None,
            copy=copy,
        )
        for graph, out_value in zip(dataset, out):
            graph[y] = out_value
        return dataset

    # noinspection PyPep8Naming
    def fit_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                              y: str = None, *,
                              X: str = None, atomic_number: str = None,
                              sample_weight: str = None,
                              copy: bool = True,
                              copy_dataset: bool = False
                              ) -> List[Dict[str, np.ndarray]]:
        r"""Fit and transform to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            y (str): Name of y information in dataset. For example "graph_properties".
            X (str): Optional X information in dataset. For example "atomic_number".
            atomic_number (list): Name of atomic number information in dataset. For example "atomic_number".
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        self.fit_dataset(dataset=dataset, y=y, X=X, atomic_number=atomic_number, sample_weight=sample_weight)
        return self.transform_dataset(
            dataset=dataset, y=y, X=X, atomic_number=atomic_number, copy=copy, copy_dataset=copy_dataset)



import unittest
import random
import typing as t

import numpy as np

from kgcnn.graph.base import GraphDict
from kgcnn.data.base import MemoryGraphDataset


class TestMemoryGraphDataset(unittest.TestCase):

    # -- UTILITY METHODS --

    def create_dataset(self,
                       num_elements: int = 10,
                       train_ratio: float = 0.8,
                       num_splits: int = 1,
                       ) -> MemoryGraphDataset:
        memory_dataset = MemoryGraphDataset()
        memory_dataset.empty(num_elements)

        indices = list(range(num_elements))
        train_indices_list = [random.sample(indices, k=int(num_elements * train_ratio))
                              for _ in range(num_splits)]
        test_indices_list = [[index for index in indices if index not in train_indices]
                             for train_indices in train_indices_list]

        for index in range(num_elements):
            graph_dict = GraphDict({
                'node_indices': np.array([0, 1, 2]),
                'node_attributes': np.array([
                    [random.random()],
                    [random.random()],
                    [random.random()],
                ]),
                'edge_indices': np.array([[0, 1], [1, 2], [2, 3]]),
                'edge_attributes': np.array([
                    [random.random()],
                    [random.random()],
                    [random.random()]
                ]),
                'graph_labels': [random.random(), random.random()],
                'train': [(i + 1)
                          for i, train_indices in enumerate(train_indices_list)
                          if index in train_indices],
                'test': [(i + 1)
                         for i, test_indices in enumerate(test_indices_list)
                         if index in test_indices],
            })
            memory_dataset[index] = graph_dict

        return memory_dataset

    # -- UNITTESTS --

    def test_creating_from_list_of_graph_dicts_basically_works(self):
        """
        If it is possible to create a MemoryGraphDataset from a list of GraphDicts
        """
        num_elements = 10

        memory_dataset = MemoryGraphDataset()
        memory_dataset.empty(num_elements)
        for i in range(num_elements):
            graph_dict = GraphDict({
                'node_indices': np.array([0, 1, 2]),
                'node_attributes': np.array([[random.random(), random.random(), random.random()]])
            })
            memory_dataset[i] = graph_dict

        self.assertIsInstance(memory_dataset, MemoryGraphDataset)
        self.assertEqual(num_elements, len(memory_dataset))

        node_indices_list = memory_dataset.obtain_property('node_indices')
        self.assertIsInstance(node_indices_list, list)
        self.assertEqual(num_elements, len(node_indices_list))

    def test_create_dataset_works(self):
        """
        If the utility method "create_dataset" of this class works as intended
        """
        num_elements = 10
        memory_dataset: MemoryGraphDataset = self.create_dataset(num_elements)
        self.assertIsInstance(memory_dataset, MemoryGraphDataset)
        self.assertEqual(num_elements, len(memory_dataset))

    def test_get_train_test_indices_works(self):
        num_elements = 10
        train_ratio = 0.8
        memory_dataset: MemoryGraphDataset = self.create_dataset(
            num_elements=num_elements,
            train_ratio=train_ratio,
            num_splits=1,
        )

        train_test_indices = memory_dataset.get_train_test_indices(
            train='train',
            test='test',
            valid=None
        )
        train_indices, test_indices = train_test_indices[0]
        self.assertIsInstance(train_indices, np.ndarray)
        self.assertEqual(round(train_ratio * num_elements), len(train_indices))

        self.assertIsInstance(test_indices, np.ndarray)
        self.assertEqual(round((1 - train_ratio) * num_elements), len(test_indices))

    def test_get_train_test_indices_with_multiple_splits_works(self):
        num_elements = 10
        train_ratio = 0.8
        num_splits = 3
        memory_dataset: MemoryGraphDataset = self.create_dataset(
            num_elements=num_elements,
            train_ratio=train_ratio,
            num_splits=num_splits
        )

        train_test_indices = memory_dataset.get_train_test_indices(
            train='train',
            test='test',
            valid=None,
            split_index=list(range(1, num_splits + 1))
        )
        self.assertEqual(num_splits, len(train_test_indices))
        for train_indices, test_indices in train_test_indices:

            self.assertIsInstance(train_indices, np.ndarray)
            self.assertEqual(round(train_ratio * num_elements), len(train_indices))

            self.assertIsInstance(test_indices, np.ndarray)
            self.assertEqual(round((1 - train_ratio) * num_elements), len(test_indices))





import tensorflow as tf
from kgcnn.layers.casting import ChangeTensorType
from kgcnn.layers.gather import GatherNodesSelection
from kgcnn.layers.modules import Dense, OptionalInputEmbedding, LazyMultiply, LazyAdd, Activation
from kgcnn.layers.pooling import PoolingLocalMessages
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.layers.relational import RelationalDense
from kgcnn.model.utils import update_model_kwargs

ks = tf.keras

# Keep track of model version from commit date in literature.
# To be updated if model is changed in a significant way.
__model_version__ = "2022.11.25"

# Implementation of GCN in `tf.keras` from paper:
# GNN-FiLM: Graph Neural Networks with Feature-wise Linear Modulation
# Marc Brockschmidt
# https://arxiv.org/abs/1906.12192


model_default = {
    "name": "GNNFilm",
    "inputs": [{"shape": (None,), "name": "node_attributes", "dtype": "float32", "ragged": True},
               {"shape": (None, 2), "name": "edge_indices", "dtype": "int64", "ragged": True},
               {"shape": (None, ), "name": "edge_relations", "dtype": "int64", "ragged": True}],
    "input_embedding": {"node": {"input_dim": 95, "output_dim": 64}},
    "dense_relation_kwargs": {"units": 64, "num_relations": 20},
    "dense_modulation_kwargs": {"units": 64, "num_relations": 20, "activation": "sigmoid"},
    "activation_kwargs": {"activation": "swish"},
    "depth": 3, "verbose": 10,
    "output_embedding": 'graph', "output_to_tensor": True,
    "output_mlp": {"use_bias": True, "units": 1,
                   "activation": "softmax"}
}


@update_model_kwargs(model_default)
def make_model(inputs: list = None,
               input_embedding: dict = None,
               depth: int = None,
               dense_relation_kwargs: dict = None,
               dense_modulation_kwargs: dict = None,
               activation_kwargs: dict = None,
               name: str = None,
               verbose: int = None,
               output_embedding: str = None,
               output_to_tensor: bool = None,
               output_mlp: dict = None
               ):
    r"""Make `GNNFilm <https://arxiv.org/abs/1906.12192>`_ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.RGCN.model_default`.

    Inputs:
        list: `[node_attributes, edge_indices, edge_relations]`

            - node_attributes (tf.RaggedTensor): Node attributes of shape `(batch, None, F)` or `(batch, None)`
              using an embedding layer.
            - edge_indices (tf.RaggedTensor): Index list for edges of shape `(batch, None, 2)`.
            - edge_relations (tf.RaggedTensor): Edge relations of shape `(batch, None)` .

    Outputs:
        tf.Tensor: Graph embeddings of shape `(batch, L)` if :obj:`output_embedding="graph"`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in :obj:`Embedding` layers.
        depth (int): Number of graph embedding units or depth of the network.
        dense_relation_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`RelationalDense` layer.
        dense_modulation_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`RelationalDense` layer.
        activation_kwargs (dict):  Dictionary of layer arguments unpacked in :obj:`Activation` layer.
        name (str): Name of the model.
        verbose (int): Level of print output.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_to_tensor (bool): Whether to cast model output to :obj:`tf.Tensor`.
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        :obj:`tf.keras.models.Model`
    """
    # Make input
    node_input = ks.layers.Input(**inputs[0])
    edge_index_input = ks.layers.Input(**inputs[1])
    edge_relations = ks.layers.Input(**inputs[2])

    # Embedding, if no feature dimension
    n = OptionalInputEmbedding(**input_embedding['node'], use_embedding=len(inputs[0]['shape']) < 2)(node_input)
    edi = edge_index_input

    # Model
    for i in range(0, depth):
        n_i, n_j = GatherNodesSelection(selection_index=[0, 1])([n, edi])
        # Note: This maybe could be done more efficiently.
        gamma = RelationalDense(**dense_modulation_kwargs)([n_i, edge_relations])
        beta = RelationalDense(**dense_modulation_kwargs)([n_i, edge_relations])
        h_j = RelationalDense(**dense_relation_kwargs)([n_j, edge_relations])
        m = LazyMultiply()([h_j, gamma])
        m = LazyAdd()([m, beta])
        h = PoolingLocalMessages(pooling_method="sum")([n, m, edi])
        n = Activation(**activation_kwargs)(h)

    # Output embedding choice
    if output_embedding == "graph":
        out = PoolingNodes()(n)
        out = MLP(**output_mlp)(out)
    elif output_embedding == "node":  # Node labeling
        out = GraphMLP(**output_mlp)(n)
        if output_to_tensor:  # For tf version < 2.8 cast to tensor below.
            out = ChangeTensorType(input_tensor_type='ragged', output_tensor_type="tensor")(out)
    else:
        raise ValueError("Unsupported output embedding for mode `GNNFilm`")

    model = ks.models.Model(inputs=[node_input, edge_index_input, edge_relations], outputs=out, name=name)
    model.__kgcnn_model_version__ = __model_version__
    return model



from kgcnn.data.visual_graph import VisualGraphDataset


class VgdMockDataset(VisualGraphDataset):
    r"""Synthetic classification dataset containing 100 small, randomly generated graphs,
    where half of them were seeded with a triangular subgraph motif,
    which is the explanation ground truth for the target class distinction.

    """

    def __init__(self):
        super(VgdMockDataset, self).__init__(name='mock')



import os.path
import numpy as np
import logging
from typing import List, Dict
from sklearn.preprocessing import StandardScaler as StandardScalerSklearn
from kgcnn.data.utils import save_json_file, load_json_file


logging.basicConfig()  # Module logger
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


class StandardScalerSklearnBase(StandardScalerSklearn):
    r"""Base standard scaler of :obj:`sklearn` with added functionality to save and load weights of this scaler similar
    to keras layers and objects.

    .. code-block:: python

        import numpy as np
        from kgcnn.scaler.scaler import StandardScalerSklearnBase
        data = np.random.rand(5).reshape((5,1))
        scaler = StandardScalerSklearnBase()
        scaler.fit(X=data)
        print(scaler.get_weights())
        print(scaler.get_config())

    """
    _attributes_list_sklearn = ["n_features_in_", "mean_", "scale_", "var_", "feature_names_in_", "n_samples_seen_"]

    def __init__(self, *, copy=True, with_mean=True, with_std=True):
        super(StandardScalerSklearnBase, self).__init__(copy=copy, with_mean=with_mean, with_std=with_std)

    def get_scaling(self):
        """Get scale of shape (1, n_properties)."""
        if not hasattr(self, "scale_"):
            return
        scale = np.array(self.scale_)
        scale = np.expand_dims(scale, axis=0)
        return scale

    def get_config(self) -> dict:
        """Get configuration for scaler."""
        config = super(StandardScalerSklearnBase, self).get_params()
        return config

    def set_config(self, config: dict):
        """Set configuration for scaler.

        Args:
            config (dict): Config dictionary.
        """
        self.set_params(**config)

    def get_weights(self) -> dict:
        """Get weights for this scaler after fit."""
        weight_dict = dict()
        for x in self._attributes_list_sklearn:
            if hasattr(self, x):
                value = getattr(self, x)
                value_update = {x: np.array(value).tolist()} if value is not None else {x: value}
                weight_dict.update(value_update)
        return weight_dict

    def set_weights(self, weights: dict):
        """Set weights for this scaler.

        Args:
            weights (dict): Weight dictionary.
        """
        for item, value in weights.items():
            if item in self._attributes_list_sklearn:
                setattr(self, item, np.array(value))
            else:
                module_logger.warning("`StandardScaler` got unknown weight '%s'." % item)

    def save_weights(self, file_path: str):
        """Save weights as numpy to file.

        Args:
            file_path: Filepath to save weights.
        """
        weights = self.get_weights()
        # Make them all numpy arrays for save.
        for key, value in weights.items():
            weights[key] = np.array(value)
        if len(weights) > 0:
            np.savez(os.path.splitext(file_path)[0] + ".npz", **weights)
        else:
            module_logger.warning("Error, no weights to save for `StandardScaler`.")

    def save(self, file_path: str):
        """Save scaler serialization to file.

        Args:
            file_path: Filepath to save scaler serialization.
        """
        conf = self.get_config()
        weights = self.get_weights()
        full_info = {"class_name": type(self).__name__, "module_name": type(self).__module__,
                     "config": conf, "weights": weights}
        save_json_file(full_info, os.path.splitext(file_path)[0] + ".json")

    def load(self, file_path: str):
        """Load scaler serialization from file.

        Args:
            file_path: Filepath to load scaler serialization.
        """
        full_info = load_json_file(file_path)
        # Could verify class_name and module_name here.
        self.set_config(full_info["config"])
        self.set_weights(full_info["weights"])
        return self

    # Similar functions that work on dataset plus property names.
    # noinspection PyPep8Naming
    def fit_dataset(self, dataset: List[Dict[str, np.ndarray]],
                    X: str = None,
                    y: str = None,
                    sample_weight: str = None):
        r"""Fit to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            X (str): Name of X information in dataset. For example "graph_properties".
            y (str): Not used.
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".

        Returns:
            self.
        """
        return self.fit(
            X=[item[X] for item in dataset],
            y=None,
            sample_weight=[item[sample_weight] for item in dataset] if sample_weight is not None else None
        )

    # noinspection PyPep8Naming
    def transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                          X: str = None,
                          copy: bool = True,
                          copy_dataset: bool = False,
                          ) -> List[Dict[str, np.ndarray]]:
        r"""Transform dataset with relevant `X` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            X (str): Name of X information in dataset. For example "graph_properties".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        if copy_dataset:
            dataset = dataset.copy()
        out = self.transform(
            X=[graph[X] for graph in dataset],
            copy=copy,
        )
        for graph, out_value in zip(dataset, out):
            graph[X] = out_value
        return dataset

    # noinspection PyPep8Naming
    def inverse_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                                  X: str = None,
                                  copy: bool = True,
                                  copy_dataset: bool = False,
                                  ) -> List[Dict[str, np.ndarray]]:
        r"""Inverse transform dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            X (str): Name of X information in dataset. For example "graph_properties".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Inverse-transformed dataset.
        """
        if copy_dataset:
            dataset = dataset.copy()
        out = self.inverse_transform(
            X=[graph[X] for graph in dataset],
            copy=copy,
        )
        for graph, out_value in zip(dataset, out):
            graph[X] = out_value
        return dataset

    # noinspection PyPep8Naming
    def fit_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                              X: str = None,
                              y: str = None,
                              sample_weight: str = None,
                              copy: bool = True,
                              copy_dataset: bool = False
                              ) -> List[Dict[str, np.ndarray]]:
        r"""Fit and transform to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            X (str): Name of X information in dataset. For example "graph_properties".
            y (str): Not used.
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        self.fit_dataset(dataset=dataset, X=X, y=y, sample_weight=sample_weight)
        return self.transform_dataset(dataset=dataset, X=X, copy=copy, copy_dataset=copy_dataset)


class StandardScaler(StandardScalerSklearnBase):
    r"""Standard scaler that inherits from :obj:`sklearn.preprocessing.StandardScaler` .
    Included unused kwarg 'atomic_number' to be compatible with some material oriented scaler.

    .. code-block:: python

        import numpy as np
        from kgcnn.scaler.scaler import StandardScaler
        data = np.random.rand(5).reshape((5,1))
        scaler = StandardScaler()
        scaler.fit(X=data)
        print(scaler.get_weights())
        print(scaler.get_config())
        print(scaler.inverse_transform(scaler.transform(X=data)))
        print(data)
        scaler.save("example.json")
        new_scaler = StandardScaler()
        new_scaler.load("example.json")
        print(new_scaler.inverse_transform(scaler.transform(X=data)))

    """

    # noinspection PyPep8Naming
    def fit(self, X, *, y=None, sample_weight=None, atomic_number=None):
        """Compute the mean and std to be used for later scaling.

        Args:
            X (np.ndarray): Array of shape (n_samples, n_features)
                The data used to compute the mean and standard deviation
                used for later scaling along the feature's axis.
            y (None): Ignored.
            sample_weight (np.ndarray): Individual weights for each sample.
            atomic_number (list, None): Ignored.

        Returns:
            self: Fitted scaler.
        """
        return super(StandardScaler, self).fit(X=X, y=y, sample_weight=sample_weight)

    # noinspection PyPep8Naming
    def partial_fit(self, X, y=None, sample_weight=None, atomic_number=None):
        r"""Online computation of mean and std on X for later scaling.
        All of X is processed as a single batch. This is intended for cases
        when :meth:`fit` is not feasible due to very large number of
        `n_samples` or because X is read from a continuous stream.
        The algorithm for incremental mean and std is given in Equation 1.5a,b
        in `Chan, et al. (1982) <https://www.tandfonline.com/doi/abs/10.1080/00031305.1983.10483115>`_ .

        Args:
            X (np.ndarray): Array of shape (n_samples, n_features)
                The data used to compute the mean and standard deviation
                used for later scaling along the feature's axis.
            y (np.ndarray, None): Ignored.
            sample_weight (np.ndarray): Array-like of shape (n_samples,), default=None
                Individual weights for each sample.
            atomic_number (list): Not used.

        Returns:
            self: Fitted scaler.
        """
        # For partial fit internally uses args and not kwargs.
        # Can not request kwargs after argument X here.
        return super(StandardScaler, self).partial_fit(X=X, y=y, sample_weight=sample_weight)

    # noinspection PyPep8Naming
    def fit_transform(self, X, *, y=None, atomic_number=None, **fit_params):
        r"""Perform fit and standardization by centering and scaling.

        Args:
            X (np.ndarray): Array of shape (n_samples, n_features).
                The data used to scale along the feature's axis.
            y (np.ndarray, None): Ignored.
            atomic_number (list): Not used.
            fit_params: Additional fit kwargs.

        Returns:
            X_tr (np.ndarray): Transformed array of shape (n_samples, n_features).
        """
        return super(StandardScaler, self).fit_transform(X=X, y=y, **fit_params)

    # noinspection PyPep8Naming
    def transform(self, X, *, copy=None, atomic_number=None):
        """Perform standardization by centering and scaling.

        Args:
            X (np.ndarray): Array of shape (n_samples, n_features).
                The data used to scale along the feature's axis.
            copy (bool): Copy the input X or not.
            atomic_number (list): Not used.

        Returns:
            X_tr (np.ndarray): Transformed array of shape (n_samples, n_features).
        """
        return super(StandardScaler, self).transform(X=X, copy=copy)

    # noinspection PyPep8Naming
    def inverse_transform(self, X, *, copy: bool = None, atomic_number=None):
        r"""Scale back the data to the original representation.

        Args:
            X (np.ndarray): Array of shape (n_samples, n_features).
                The data used to scale along the feature's axis.
            copy (bool): Copy the input X or not.
            atomic_number (list): Not used.

        Returns:
            X_tr (np.ndarray): Transformed array of shape (n_samples, n_features).
        """
        return super(StandardScaler, self).inverse_transform(X=X, copy=copy)

    # Similar functions that work on dataset plus property names.
    # noinspection PyPep8Naming
    def fit_dataset(self, dataset: List[Dict[str, np.ndarray]],
                    X: str = None, *,
                    y: str = None,
                    atomic_number: str = None, sample_weight: str = None):
        r"""Fit to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            X (str): Name of X information in dataset. For example "graph_properties".
            y (str): Not used.
            atomic_number (str): Not used.
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".

        Returns:
            self.
        """
        return super(StandardScaler, self).fit_dataset(dataset=dataset, X=X, y=y, sample_weight=sample_weight)

    # noinspection PyPep8Naming
    def transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                          X: str = None, *,
                          atomic_number: str = None,
                          copy: bool = True,
                          copy_dataset: bool = False,
                          ) -> List[Dict[str, np.ndarray]]:
        r"""Transform dataset with relevant `X` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            X (str): Name of X information in dataset. For example "graph_properties".
            atomic_number (str): Not used.
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        return super(StandardScaler, self).transform_dataset(dataset=dataset, X=X, copy_dataset=copy_dataset, copy=copy)

    # noinspection PyPep8Naming
    def inverse_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                                  X: str = None, *,
                                  atomic_number: str = None,
                                  copy: bool = True,
                                  copy_dataset: bool = False,
                                  ) -> List[Dict[str, np.ndarray]]:
        r"""Inverse transform dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            X (str): Name of X information in dataset. For example "graph_properties".
            atomic_number (str): Not used.
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Inverse-transformed dataset.
        """
        return super(StandardScaler, self).inverse_transform_dataset(
            dataset=dataset, X=X, copy_dataset=copy_dataset, copy=copy)

    # noinspection PyPep8Naming
    def fit_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                              X: str = None, *,
                              y: str = None,
                              atomic_number: str = None,
                              sample_weight: str = None,
                              copy: bool = True,
                              copy_dataset: bool = False
                              ) -> List[Dict[str, np.ndarray]]:
        r"""Fit and transform to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            X (str): Name of X information in dataset. For example "graph_properties".
            y (str): Not used.
            atomic_number (str): Not used.
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        self.fit_dataset(dataset=dataset, X=X, y=y, sample_weight=sample_weight)
        return self.transform_dataset(dataset=dataset, X=X, copy=copy, copy_dataset=copy_dataset)


class StandardLabelScaler(StandardScalerSklearnBase):
    r"""Standard scaler for labels that inherits from :obj:`sklearn.preprocessing.StandardScaler` .
    Included unused kwarg 'atomic_number' to be compatible with some material oriented scaler.
    Uses `y` argument for scaling labels and `X` is ignored.

    .. code-block:: python

        import numpy as np
        from kgcnn.data.transform.scaler.scaler import StandardLabelScaler
        data = np.random.rand(5).reshape((5,1))
        scaler = StandardLabelScaler()
        scaler.fit(y=data)
        print(scaler.fit_transform(y=data))
        print(scaler.get_weights())
        print(scaler.get_config())
        print(scaler.inverse_transform(y=scaler.transform(y=data)))
        print(data)
        scaler.save("example.json")
        new_scaler = StandardLabelScaler()
        new_scaler.load("example.json")
        print(new_scaler.inverse_transform(y=scaler.transform(y=data)))

    """

    def _validate_input(self, y, x):
        if x is not None and y is None:
            raise ValueError(
                "Got X but y is 'None' for `%s`. Report this issue if sklearn API change. " % type(self).__name__)
        if y is None and x is None:
            raise ValueError(
                "Require labels as `y` for `%s`. Input must be e.g. 'fit(data)'." % type(self).__name__)

    # noinspection PyPep8Naming
    def fit(self, y: np.ndarray, *, X=None, sample_weight=None, atomic_number=None):
        r"""Compute the mean and std to be used for later scaling.

        Args:
            y (np.ndarray): Array of shape (n_samples, n_labels)
                The data used to compute the mean and standard deviation
                used for later scaling along the feature's axis.
            X (None): Ignored.
            sample_weight (np.ndarray): Individual weights for each sample.
            atomic_number (list): Ignored.

        Returns:
            self: Fitted scaler.
        """
        # fit() of sklearn uses reset and partial fit. Just adding y in place of X.
        self._validate_input(y, X)
        return super(StandardLabelScaler, self).fit(y, sample_weight=sample_weight)

    # noinspection PyPep8Naming
    def partial_fit(self, y: np.ndarray, X=None, sample_weight=None, atomic_number=None):
        r"""Online computation of mean and std on y for later scaling.
        All of y is processed as a single batch. This is intended for cases
        when :meth:`fit` is not feasible due to very large number of
        `n_samples` or because y is read from a continuous stream.
        The algorithm for incremental mean and std is given in Equation 1.5a,b
        in `Chan, et al. (1982) <https://www.tandfonline.com/doi/abs/10.1080/00031305.1983.10483115>`_ .

        Args:
            y (np.ndarray): Array of shape (n_samples, n_labels)
                The data used to compute the mean and standard deviation
                used for later scaling along the feature's axis.
            X (None): Ignored.
            sample_weight (np.ndarray): Individual weights for each sample.
            atomic_number (list): Ignored.

        Returns:
            self: Fitted scaler.
        """
        # For partial fit internally uses args and not kwargs.
        # Can not request kwargs after argument X, y here.
        self._validate_input(y, X)
        return super(StandardLabelScaler, self).partial_fit(y, sample_weight=sample_weight)

    # noinspection PyPep8Naming
    def fit_transform(self, y: np.ndarray, *, X=None, atomic_number=None, **fit_params):
        r"""Perform fit and standardization by centering and scaling.

        Args:
            y (np.ndarray): Array of shape (n_samples, n_labels)
                The data used to compute the mean and standard deviation
                used for later scaling along the feature's axis.
            X (None): Ignored.
            atomic_number (list): Ignored.
            fit_params (Any): Kwargs for fit.

        Returns:
            y_tr (np.ndarray): Transformed array of shape (n_samples, n_labels).
        """
        return super(StandardLabelScaler, self).fit_transform(y, **fit_params)

    # noinspection PyPep8Naming
    def transform(self, y: np.ndarray, *, X=None, copy=None, atomic_number=None):
        r"""Perform standardization by centering and scaling.

        Args:
            y (None): Array of shape (n_samples, n_labels)
                The data used to scale along the feature's axis.
            X (None): Ignored.
            atomic_number (list): Ignored.
            copy (bool): Copy the input `y` or not.

        Returns:
            y_tr (np.ndarray): Transformed array of shape (n_samples, n_labels).
        """
        # Just changing order of x,y here.
        return super(StandardLabelScaler, self).transform(y, copy=copy)

    # noinspection PyPep8Naming
    def inverse_transform(self, y: np.ndarray = None, *, X=None, copy: bool = None, atomic_number=None):
        r"""Scale back the data to the original representation.

        Args:
            y (None): Array of shape (n_samples, n_labels)
                The data used to scale along the feature's axis.
            X (np.ndarray, None): Ignored. Default is None.
            atomic_number (list): Ignored.
            copy (bool): Copy the input `y` or not.

        Returns:
            y_tr (np.ndarray): Transformed array of shape (n_samples, n_labels).
        """
        # Just changing order of x,y here.
        return super(StandardLabelScaler, self).inverse_transform(y, copy=copy)

    # Similar functions that work on dataset plus property names.
    # noinspection PyPep8Naming
    def fit_dataset(self, dataset: List[Dict[str, np.ndarray]],
                    y: str = None, *,
                    X: str = None,
                    atomic_number: str = None,
                    sample_weight: str = None):
        r"""Fit to dataset with relevant `X` , `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            y (str): Name of y information in dataset. For example "graph_labels".
            X (str): Not used.
            atomic_number (str): Not used.
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".

        Returns:
            self.
        """
        # Just changing order of x,y here.
        return super(StandardLabelScaler, self).fit_dataset(dataset=dataset, X=y, y=None, sample_weight=sample_weight)

    # noinspection PyPep8Naming
    def transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                          y: str = None, *,
                          X: str = None,
                          atomic_number: str = None,
                          copy: bool = True,
                          copy_dataset: bool = False,
                          ) -> List[Dict[str, np.ndarray]]:
        r"""Transform dataset with relevant `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            y (str): Name of y information in dataset. For example "graph_labels".
            X (str): Not used.
            atomic_number (str): Not used.
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        # Just changing order of x,y here.
        return super(StandardLabelScaler, self).transform_dataset(
            dataset=dataset, X=y, copy_dataset=copy_dataset, copy=copy)

    # noinspection PyPep8Naming
    def inverse_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                                  y: str = None, *,
                                  X: str = None,
                                  atomic_number: str = None,
                                  copy: bool = True,
                                  copy_dataset: bool = False,
                                  ) -> List[Dict[str, np.ndarray]]:
        r"""Inverse transform dataset with relevant `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            y (str): Name of y information in dataset. For example "graph_labels".
            X (str): Not used.
            atomic_number (str): Not used.
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Inverse-transformed dataset.
        """
        return super(StandardLabelScaler, self).inverse_transform_dataset(
            dataset=dataset, X=y, copy_dataset=copy_dataset, copy=copy)

    # noinspection PyPep8Naming
    def fit_transform_dataset(self, dataset: List[Dict[str, np.ndarray]],
                              y: str = None, *,
                              X: str = None,
                              atomic_number: str = None,
                              sample_weight: str = None,
                              copy: bool = True,
                              copy_dataset: bool = False
                              ) -> List[Dict[str, np.ndarray]]:
        r"""Fit and transform to dataset with relevant `y` information.

        Args:
            dataset (list): Dataset of type `List[Dict]` with dictionary of numpy arrays.
            y (str): Name of y information in dataset. For example "graph_properties".
            X (str): Not used.
            atomic_number (str): Not used.
            sample_weight (str): Name of sample weight information in dataset. For example "sample_weight".
            copy (bool): Whether to copy data for transformation. Default is True.
            copy_dataset (bool): Whether to copy full dataset. Default is False.

        Returns:
            dataset: Transformed dataset.
        """
        # Uses the functions of this class.
        self.fit_dataset(dataset=dataset, y=y, X=X, atomic_number=atomic_number, sample_weight=sample_weight)
        return self.transform_dataset(
            dataset=dataset, y=y, X=X, atomic_number=atomic_number, copy=copy, copy_dataset=copy_dataset)



import ase
import tensorflow as tf
import numpy as np
from ase import Atoms
from ase.calculators.calculator import Calculator
from kgcnn.graph.base import GraphDict
from kgcnn.data.base import MemoryGraphList
from copy import deepcopy
from typing import Union, List

ks = tf.keras


class AtomsToGraphConverter:
    r"""Convert :obj:`ase.Atoms` object to a :obj:`GraphDict` dictionary.

    Simple tool to get named properties from :obj:`ase.Atoms`. Note that the actual graph indices and connections
    have to be generated with :obj:`GraphPreProcessorBase` instances.

    Example usage:

     .. code-block:: python

        import numpy as np
        from ase import Atoms
        from kgcnn.md.ase_calc import AtomsToGraphConverter
        atoms = Atoms('N2', positions=[[0, 0, -1], [0, 0, 1]])
        trans = AtomsToGraphConverter({
            "node_number": "get_atomic_numbers", "node_coordinates": "get_positions",
            "node_symbol": "get_chemical_symbols"})
        print(trans(atoms))

    """

    def __init__(self, properties: dict = None):
        r"""Set up :obj:`AtomsToGraphConverter` converter.

        Args:
            properties (dict): Dictionary of graph properties linked to :obj:`ase.Atoms` get attribute methods.
                Default is {"node_number": "get_atomic_numbers", "node_coordinates": "get_positions",
                "node_symbol": "get_chemical_symbols"}.
        """
        if properties is None:
            properties = {"node_number": "get_atomic_numbers", "node_coordinates": "get_positions",
                          "node_symbol": "get_chemical_symbols"}
        self.properties = deepcopy(properties)

    def __call__(self, atoms: Union[List[Atoms], Atoms]) -> MemoryGraphList:
        r"""Make :obj:`GraphDict` objects from :obj:`ase.Atoms`.

        Args:
            atoms (list): List of :obj:`ase.Atoms` objects or single ASE atoms object.

        Returns:
            MemoryGraphList: List of :obj:`GraphDict` objects.
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]

        graph_list = []
        for i, x in enumerate(atoms):
            g = GraphDict()
            for key, value in self.properties.items():
                g.update({key: np.array(getattr(Atoms, value)(x))})
            graph_list.append(g)

        return MemoryGraphList(graph_list)

    def get_config(self):
        """Get config for this class."""
        config = {"properties": self.properties}
        return config


class KgcnnSingleCalculator(ase.calculators.calculator.Calculator):
    r"""ASE calculator for machine learning models from :obj:`kgcnn`."""

    implemented_properties = ["energy", "forces"]

    def __init__(self,
                 model_predictor=None,
                 atoms_converter: AtomsToGraphConverter = None,
                 squeeze_energy: bool = True,
                 **kwargs):
        super(KgcnnSingleCalculator, self).__init__(**kwargs)
        self.model_predictor = model_predictor
        self.atoms_converter = atoms_converter
        self.squeeze_energy = squeeze_energy

    # Interface to ASE calculator scheme.
    def calculate(self, atoms=None, properties=None, system_changes=None):

        if not self.calculation_required(atoms, properties):
            # Nothing to do.
            return
        super(KgcnnSingleCalculator, self).calculate(atoms=atoms, properties=properties, system_changes=system_changes)

        graph_list = self.atoms_converter(atoms)
        output_dict = self.model_predictor(graph_list)

        # Update.
        assert len(output_dict) == 1, "ASE Calculator updates only one structure for now."
        self.results.update(output_dict[0].to_dict())
        if self.squeeze_energy:
            # For single energy only.
            if len(self.results["energy"].shape) > 0:
                self.results["energy"] = np.squeeze(self.results["energy"])



import unittest

import tensorflow as tf

from kgcnn.xai.testing import MockContext
from kgcnn.literature.GNNExplain import GnnExplainer


class TestGnnExplainer(unittest.TestCase):

    def test_basically_works(self):
        num_targets = 1
        with MockContext(num_targets=num_targets) as mock:
            gnn_explainer = GnnExplainer(
                channels=num_targets,
                verbose=True
            )
            node_importances, edge_importances = gnn_explainer(
                model=mock.model,
                x=mock.x,
                y=mock.y,
            )
            assert isinstance(node_importances, tf.RaggedTensor)
            assert isinstance(edge_importances, tf.RaggedTensor)

    def test_multiple_targets_works(self):
        num_targets = 2
        with MockContext(num_targets=num_targets) as mock:
            gnn_explainer = GnnExplainer(
                channels=num_targets,
                verbose=True
            )
            node_importances, edge_importances = gnn_explainer(
                model=mock.model,
                x=mock.x,
                y=mock.y,
            )
            assert isinstance(node_importances, tf.RaggedTensor)
            assert isinstance(edge_importances, tf.RaggedTensor)





import tensorflow as tf
from kgcnn.layers.casting import ChangeTensorType
from kgcnn.layers.modules import LazyConcatenate
from kgcnn.layers.conv.wacsf_conv import wACSFAng, wACSFRad
from kgcnn.layers.conv.acsf_conv import ACSFG2, ACSFG4, ACSFConstNormalization
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.model.utils import update_model_kwargs
from kgcnn.layers.mlp import RelationalMLP
from kgcnn.layers.norm import GraphBatchNormalization

ks = tf.keras

# Keep track of model version from commit date in literature.
# To be updated if model is changed in a significant way.
__model_version__ = "2023.01.17"

# Implementation of HDNNP in `tf.keras` from paper:
# Atom-centered symmetry functions for constructing high-dimensional neural network potentials
# by Jörg Behler (2011)
# https://aip.scitation.org/doi/abs/10.1063/1.3553717


model_default_weighted = {
    "name": "HDNNP2nd",
    "inputs": [{"shape": (None,), "name": "node_number", "dtype": "int64", "ragged": True},
               {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32", "ragged": True},
               {"shape": (None, 2), "name": "edge_indices", "dtype": "int64", "ragged": True},
               {"shape": (None, 3), "name": "angle_indices_nodes", "dtype": "int64", "ragged": True}],
    "w_acsf_ang_kwargs": {},
    "w_acsf_rad_kwargs": {},
    "normalize_kwargs": None,
    "const_normalize_kwargs": None,
    "mlp_kwargs": {"units": [64, 64, 64],
                   "num_relations": 96,
                   "activation": ["swish", "swish", "linear"]},
    "node_pooling_args": {"pooling_method": "sum"},
    "verbose": 10,
    "output_embedding": "graph", "output_to_tensor": True,
    "use_output_mlp": False,
    "output_mlp": {"use_bias": [True, True], "units": [64, 1],
                   "activation": ["swish", "linear"]}
}


@update_model_kwargs(model_default_weighted)
def make_model_weighted(inputs: list = None,
                        node_pooling_args: dict = None,
                        name: str = None,
                        verbose: int = None,
                        w_acsf_ang_kwargs: dict = None,
                        w_acsf_rad_kwargs: dict = None,
                        normalize_kwargs: dict = None,
                        const_normalize_kwargs: dict = None,
                        mlp_kwargs: dict = None,
                        output_embedding: str = None,
                        use_output_mlp: bool = None,
                        output_to_tensor: bool = None,
                        output_mlp: dict = None
                        ):
    r"""Make 2nd generation `HDNNP <https://arxiv.org/abs/1706.08566>`_ graph network via functional API.

    Default parameters can be found in :obj:`kgcnn.literature.HDNNP2nd.model_default_weighted`.
    Uses weighted `wACSF <https://arxiv.org/abs/1712.05861>`_ .

    Inputs:
        list: `[node_number, node_coordinates, edge_indices, angle_indices_nodes]`

            - node_number (tf.RaggedTensor): Atomic number of shape `(batch, None)` .
            - node_coordinates (tf.RaggedTensor): Node (atomic) coordinates of shape `(batch, None, 3)`.
            - edge_indices (tf.RaggedTensor): Index list for edges of shape `(batch, None, 2)`.
            - angle_indices_nodes (tf.RaggedTensor): Index list for angles of shape `(batch, None, 3)`.

    Outputs:
        tf.Tensor: Graph embeddings of shape `(batch, L)` if :obj:`output_embedding="graph"`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        node_pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layers.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        w_acsf_ang_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`wACSFAng` layer.
        w_acsf_rad_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`wACSFRad` layer.
        mlp_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`RelationalMLP` layer.
        normalize_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`GraphBatchNormalization` layer.
        const_normalize_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`ACSFConstNormalization` layer.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        use_output_mlp (bool): Whether to use the final output MLP. Possibility to skip final MLP.
        output_to_tensor (bool): Whether to cast model output to :obj:`tf.Tensor`.
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        :obj:`tf.keras.models.Model`
    """
    # Make input
    node_input = ks.layers.Input(**inputs[0])
    xyz_input = ks.layers.Input(**inputs[1])
    edge_index_input = ks.layers.Input(**inputs[2])
    angle_index_input = ks.layers.Input(**inputs[3])

    # ACSF representation.
    rep_rad = wACSFRad(**w_acsf_rad_kwargs)([node_input, xyz_input, edge_index_input])
    rep_ang = wACSFAng(**w_acsf_ang_kwargs)([node_input, xyz_input, angle_index_input])
    rep = LazyConcatenate()([rep_rad, rep_ang])

    # Normalization
    if normalize_kwargs:
        rep = GraphBatchNormalization(**normalize_kwargs)(rep)
    if const_normalize_kwargs:
        rep = ACSFConstNormalization(**const_normalize_kwargs)(rep)

    # learnable NN.
    n = RelationalMLP(**mlp_kwargs)([rep, node_input])

    # Output embedding choice
    if output_embedding == 'graph':
        out = PoolingNodes(**node_pooling_args)(n)
        if use_output_mlp:
            out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = n
        if use_output_mlp:
            out = GraphMLP(**output_mlp)(out)
        if output_to_tensor:  # For tf version < 2.8 cast to tensor below.
            out = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="tensor")(out)
    else:
        raise ValueError("Unsupported output embedding for mode `HDNNP2nd`")

    model = ks.models.Model(
        inputs=[node_input, xyz_input, edge_index_input, angle_index_input], outputs=out, name=name)
    model.__kgcnn_model_version__ = __model_version__
    return model


model_default_behler = {
    "name": "HDNNP2nd",
    "inputs": [{"shape": (None,), "name": "node_number", "dtype": "int64", "ragged": True},
               {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32", "ragged": True},
               {"shape": (None, 2), "name": "edge_indices", "dtype": "int64", "ragged": True},
               {"shape": (None, 3), "name": "angle_indices_nodes", "dtype": "int64", "ragged": True}],
    "g2_kwargs": {"eta": [0.0, 0.3], "rs": [0.0, 3.0], "rc": 10.0, "elements": [1, 6, 16]},
    "g4_kwargs": {"eta": [0.0, 0.3], "lamda": [-1.0, 1.0], "rc": 6.0,
                  "zeta": [1.0, 8.0], "elements": [1, 6, 16], "multiplicity": 2.0},
    "normalize_kwargs": {},
    "const_normalize_kwargs": None,
    "mlp_kwargs": {"units": [64, 64, 64],
                   "num_relations": 96,
                   "activation": ["swish", "swish", "linear"]},
    "node_pooling_args": {"pooling_method": "sum"},
    "verbose": 10,
    "output_embedding": "graph", "output_to_tensor": True,
    "use_output_mlp": False,
    "output_mlp": {"use_bias": [True, True], "units": [64, 1],
                   "activation": ["swish", "linear"]}
}


@update_model_kwargs(model_default_behler)
def make_model_behler(inputs: list = None,
                      node_pooling_args: dict = None,
                      name: str = None,
                      verbose: int = None,
                      normalize_kwargs: dict = None,
                      const_normalize_kwargs: dict = None,
                      g2_kwargs: dict = None,
                      g4_kwargs: dict = None,
                      mlp_kwargs: dict = None,
                      output_embedding: str = None,
                      use_output_mlp: bool = None,
                      output_to_tensor: bool = None,
                      output_mlp: dict = None
                      ):
    r"""Make 2nd generation `HDNNP <https://arxiv.org/abs/1706.08566>`_ graph network via functional API.

    Default parameters can be found in :obj:`kgcnn.literature.HDNNP2nd.model_default_behler`.

    Inputs:
        list: `[node_number, node_coordinates, edge_indices, angle_indices_nodes]`

            - node_number (tf.RaggedTensor): Atomic number of shape `(batch, None)` .
            - node_coordinates (tf.RaggedTensor): Node (atomic) coordinates of shape `(batch, None, 3)`.
            - edge_indices (tf.RaggedTensor): Index list for edges of shape `(batch, None, 2)`.
            - angle_indices_nodes (tf.RaggedTensor): Index list for angles of shape `(batch, None, 3)`.

    Outputs:
        tf.Tensor: Graph embeddings of shape `(batch, L)` if :obj:`output_embedding="graph"`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        node_pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layers.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        g2_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`ACSFG2` layer.
        g4_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`ACSFG4` layer.
        normalize_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`GraphBatchNormalization` layer.
        const_normalize_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`ACSFConstNormalization` layer.
        mlp_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`RelationalMLP` layer.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        use_output_mlp (bool): Whether to use the final output MLP. Possibility to skip final MLP.
        output_to_tensor (bool): Whether to cast model output to :obj:`tf.Tensor`.
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        :obj:`tf.keras.models.Model`
    """
    # Make input
    node_input = ks.layers.Input(**inputs[0])
    xyz_input = ks.layers.Input(**inputs[1])
    edge_index_input = ks.layers.Input(**inputs[2])
    angle_index_input = ks.layers.Input(**inputs[3])

    # ACSF representation.
    rep_g2 = ACSFG2(**ACSFG2.make_param_table(**g2_kwargs))([node_input, xyz_input, edge_index_input])
    rep_g4 = ACSFG4(**ACSFG4.make_param_table(**g4_kwargs))([node_input, xyz_input, angle_index_input])
    rep = LazyConcatenate()([rep_g2, rep_g4])

    # Normalization
    if normalize_kwargs:
        rep = GraphBatchNormalization(**normalize_kwargs)(rep)
    if const_normalize_kwargs:
        rep = ACSFConstNormalization(**const_normalize_kwargs)(rep)

    # learnable NN.
    n = RelationalMLP(**mlp_kwargs)([rep, node_input])

    # Output embedding choice
    if output_embedding == 'graph':
        out = PoolingNodes(**node_pooling_args)(n)
        if use_output_mlp:
            out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = n
        if use_output_mlp:
            out = GraphMLP(**output_mlp)(out)
        if output_to_tensor:  # For tf version < 2.8 cast to tensor below.
            out = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="tensor")(out)
    else:
        raise ValueError("Unsupported output embedding for mode `HDNNP2nd`")

    model = ks.models.Model(
        inputs=[node_input, xyz_input, edge_index_input, angle_index_input], outputs=out, name=name)

    model.__kgcnn_model_version__ = __model_version__
    return model


model_atom_wise_default = {
    "name": "HDNNP2nd",
    "inputs": [{"shape": (None,), "name": "node_number", "dtype": "int64", "ragged": True},
               {"shape": (None, 3), "name": "node_representation", "dtype": "float32", "ragged": True}],
    "mlp_kwargs": {"units": [64, 64, 64],
                   "num_relations": 96,
                   "activation": ["swish", "swish", "linear"]},
    "node_pooling_args": {"pooling_method": "sum"},
    "verbose": 10,
    "output_embedding": "graph", "output_to_tensor": True,
    "use_output_mlp": False,
    "output_mlp": {"use_bias": [True, True], "units": [64, 1],
                   "activation": ["swish", "linear"]}
}


@update_model_kwargs(model_atom_wise_default)
def make_model_atom_wise(inputs: list = None,
                         node_pooling_args: dict = None,
                         name: str = None,
                         verbose: int = None,
                         mlp_kwargs: dict = None,
                         output_embedding: str = None,
                         use_output_mlp: bool = None,
                         output_to_tensor: bool = None,
                         output_mlp: dict = None
                         ):
    r"""Make 2nd generation `HDNNP <https://arxiv.org/abs/1706.08566>`_ network via functional API.

    Default parameters can be found in :obj:`kgcnn.literature.HDNNP2nd.model_atom_wise_default`.

    Inputs:
        list: `[node_number, node_representation]`

            - node_number (tf.RaggedTensor): Atomic number of shape `(batch, None)` .
            - node_representation (tf.RaggedTensor): Node (atomic) features of shape `(batch, None, F)`

    Outputs:
        tf.Tensor: Graph embeddings of shape `(batch, L)` if :obj:`output_embedding="graph"`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        node_pooling_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layers.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        mlp_kwargs (dict): Dictionary of layer arguments unpacked in :obj:`RelationalMLP` layer.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        use_output_mlp (bool): Whether to use the final output MLP. Possibility to skip final MLP.
        output_to_tensor (bool): Whether to cast model output to :obj:`tf.Tensor`.
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        :obj:`tf.keras.models.Model`
    """
    # Make input
    node_input = ks.layers.Input(**inputs[0])
    rep_input = ks.layers.Input(**inputs[1])

    # learnable NN.
    n = RelationalMLP(**mlp_kwargs)([rep_input, node_input])

    # Output embedding choice
    if output_embedding == 'graph':
        out = PoolingNodes(**node_pooling_args)(n)
        if use_output_mlp:
            out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = n
        if use_output_mlp:
            out = GraphMLP(**output_mlp)(out)
        if output_to_tensor:  # For tf version < 2.8 cast to tensor below.
            out = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="tensor")(out)
    else:
        raise ValueError("Unsupported output embedding for mode `HDNNP2nd`")

    model = ks.models.Model(
        inputs=[node_input, rep_input], outputs=out, name=name)

    model.__kgcnn_model_version__ = __model_version__
    return model


# For default, the weighted ACSF are used, since they do should in principle work for all elements.
make_model = make_model_weighted



import tensorflow as tf
from kgcnn.layers.casting import ChangeTensorType
from kgcnn.layers.conv.gat_conv import AttentionHeadGAT
from kgcnn.layers.modules import LazyConcatenate, Dense, LazyAverage, Activation, \
    OptionalInputEmbedding
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.model.utils import update_model_kwargs

ks = tf.keras

# Keep track of model version from commit date in literature.
# To be updated if model is changed in a significant way.
__model_version__ = "2022.11.25"

# Implementation of GAT in `tf.keras` from paper:
# Graph Attention Networks
# by Petar Veličković, Guillem Cucurull, Arantxa Casanova, Adriana Romero, Pietro Liò, Yoshua Bengio (2018)
# https://arxiv.org/abs/1710.10903

model_default = {
    "name": "GAT",
    "inputs": [{"shape": (None,), "name": "node_attributes", "dtype": "float32", "ragged": True},
               {"shape": (None,), "name": "edge_attributes", "dtype": "float32", "ragged": True},
               {"shape": (None, 2), "name": "edge_indices", "dtype": "int64", "ragged": True}],
    "input_embedding": {"node": {"input_dim": 95, "output_dim": 64},
                        "edge": {"input_dim": 5, "output_dim": 64}},
    "attention_args": {"units": 32, "use_final_activation": False, "use_edge_features": True,
                       "has_self_loops": True, "activation": "kgcnn>leaky_relu", "use_bias": True},
    "pooling_nodes_args": {"pooling_method": "mean"},
    "depth": 3, "attention_heads_num": 5,
    "attention_heads_concat": False, "verbose": 10,
    "output_embedding": "graph", "output_to_tensor": True,
    "output_mlp": {"use_bias": [True, True, False], "units": [25, 10, 1],
                   "activation": ["relu", "relu", "sigmoid"]}
}


@update_model_kwargs(model_default)
def make_model(inputs: list = None,
               input_embedding: dict = None,
               attention_args: dict = None,
               pooling_nodes_args: dict = None,
               depth: int = None,
               attention_heads_num: int = None,
               attention_heads_concat: bool = None,
               name: str = None,
               verbose: int = None,
               output_embedding: str = None,
               output_to_tensor: bool = None,
               output_mlp: dict = None
               ):
    r"""Make `GAT <https://arxiv.org/abs/1710.10903>`_ graph network via functional API.
    Default parameters can be found in :obj:`kgcnn.literature.GAT.model_default`.

    Inputs:
        list: `[node_attributes, edge_attributes, edge_indices]`

            - node_attributes (tf.RaggedTensor): Node attributes of shape `(batch, None, F)` or `(batch, None)`
              using an embedding layer.
            - edge_attributes (tf.RaggedTensor): Edge attributes of shape `(batch, None, F)` or `(batch, None)`
              using an embedding layer.
            - edge_indices (tf.RaggedTensor): Index list for edges of shape `(batch, None, 2)`.

    Outputs:
        tf.Tensor: Graph embeddings of shape `(batch, L)` if :obj:`output_embedding="graph"`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in :obj:`Embedding` layers.
        attention_args (dict): Dictionary of layer arguments unpacked in :obj:`AttentionHeadGAT` layer.
        pooling_nodes_args (dict): Dictionary of layer arguments unpacked in :obj:`PoolingNodes` layer.
        depth (int): Number of graph embedding units or depth of the network.
        attention_heads_num (int): Number of attention heads to use.
        attention_heads_concat (bool): Whether to concat attention heads, or simply average heads.
        name (str): Name of the model.
        verbose (int): Level of print output.
        output_embedding (str): Main embedding task for graph network. Either "node", "edge" or "graph".
        output_to_tensor (bool): Whether to cast model output to :obj:`tf.Tensor`.
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification :obj:`MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        :obj:`tf.keras.models.Model`
    """

    # Make input
    node_input = ks.layers.Input(**inputs[0])
    edge_input = ks.layers.Input(**inputs[1])
    edge_index_input = ks.layers.Input(**inputs[2])

    # Embedding, if no feature dimension
    n = OptionalInputEmbedding(**input_embedding['node'],
                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
    ed = OptionalInputEmbedding(**input_embedding['edge'],
                                use_embedding=len(inputs[1]['shape']) < 2)(edge_input)
    edi = edge_index_input

    # Model
    nk = Dense(units=attention_args["units"], activation="linear")(n)
    for i in range(0, depth):
        heads = [AttentionHeadGAT(**attention_args)([nk, ed, edi]) for _ in range(attention_heads_num)]
        if attention_heads_concat:
            nk = LazyConcatenate(axis=-1)(heads)
        else:
            nk = LazyAverage()(heads)
            nk = Activation(activation=attention_args["activation"])(nk)
    n = nk

    # Output embedding choice
    if output_embedding == 'graph':
        out = PoolingNodes(**pooling_nodes_args)(n)
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = GraphMLP(**output_mlp)(n)
        if output_to_tensor:  # For tf version < 2.8 cast to tensor below.
            out = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="tensor")(out)
    else:
        raise ValueError("Unsupported output embedding for `GAT`")

    model = ks.models.Model(inputs=[node_input, edge_input, edge_index_input], outputs=out, name=name)
    model.__kgcnn_model_version__ = __model_version__
    return model



from kgcnn.data.datasets.MoleculeNetDataset2018 import MoleculeNetDataset2018


class LipopDataset(MoleculeNetDataset2018):
    r"""Store and process 'Lipop' dataset from `MoleculeNet <https://moleculenet.org/>`_ database.
    Class inherits from :obj:`MoleculeNetDataset2018` and downloads dataset on class initialization.
    Compare reference:
    `DeepChem <https://deepchem.readthedocs.io/en/latest/api_reference/moleculenet.html#lipo-datasets>`__ reading:
    Experimental results of octanol/water distribution coefficient(logD at pH 7.4).
    Description in DeepChem reads: 'Lipophilicity is an important feature of drug molecules that affects both
    membrane permeability and solubility. The lipophilicity dataset, curated from ChEMBL database, provides
    experimental results of octanol/water distribution coefficient (logD at pH 7.4) of 4200 compounds.'
    Random or Scaffold splitting is recommended for this dataset.

    References:

        (1) Hersey, A. ChEMBL Deposited Data Set - AZ dataset; 2015. `<https://doi.org/10.6019/chembl3301361>`_

    """

    def __init__(self, reload=False, verbose: int = 10):
        r"""Initialize Lipop dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        super(LipopDataset, self).__init__("Lipop", reload=reload, verbose=verbose)



from kgcnn.data.datasets.MatBenchDataset2020 import MatBenchDataset2020


class MatProjectEFormDataset(MatBenchDataset2020):
    r"""Store and process :obj:`MatProjectEFormDataset` from `MatBench <https://matbench.materialsproject.org/>`__
    database.  Name within Matbench: 'matbench_mp_e_form'.

    Matbench test dataset for predicting DFT formation energy from structure.
    Adapted from Materials Project database. Removed entries having formation energy more than 2.5eV and those
    containing noble gases. Retrieved April 2, 2019. For benchmarking w/ nested cross validation,
    the order of the dataset must be identical to the retrieved data; refer to the Automatminer/Matbench publication
    for more details.

        * Number of samples: 132752.
        * Task type: regression.
        * Input type: structure.

    """

    def __init__(self, reload=False, verbose: int = 10):
        r"""Initialize 'matbench_mp_e_form' dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        # Use default base class init()
        super(MatProjectEFormDataset, self).__init__("matbench_mp_e_form", reload=reload, verbose=verbose)
        self.label_names = "e_form"
        self.label_units = "eV/atom"



from kgcnn.data.datasets.MatBenchDataset2020 import MatBenchDataset2020


class MatProjectPhononsDataset(MatBenchDataset2020):
    """Store and process :obj:`MatProjectPhononsDataset` from `MatBench <https://matbench.materialsproject.org/>`__
    database. Name within Matbench: 'matbench_phonons'.

    Matbench test dataset for predicting vibration properties from crystal structure. Original data retrieved
    from Petretto et al. Original calculations done via ABINIT in the harmonic approximation based on density
    functional perturbation theory. Removed entries having a formation energy (or energy above the convex hull)
    more than 150meV. For benchmarking w/ nested cross validation, the order of the dataset must be identical to the
    retrieved data; refer to the Automatminer/Matbench publication for more details.

        * Number of samples: 1265
        * Task type: regression
        * Input type: structure

    last phdos peak: Target variable. Frequency of the highest frequency optical phonon mode peak, in units of 1/cm;
    may be used as an estimation of dominant longitudinal optical phonon frequency.

    """

    def __init__(self, reload=False, verbose: int = 10):
        r"""Initialize 'matbench_mp_e_form' dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        # Use default base class init()
        super(MatProjectPhononsDataset, self).__init__("matbench_phonons", reload=reload, verbose=verbose)
        self.label_names = "omega_max"
        self.label_units = "1/cm"



import time
import tensorflow as tf
import numpy as np
from typing import Union, List, Callable, Dict
from kgcnn.data.base import MemoryGraphList
from kgcnn.graph.base import GraphDict
from kgcnn.utils.serial import deserialize, serialize

ks = tf.keras


class MolDynamicsModelPredictor:
    r"""Model predictor class that adds pre- and postprocessors to the keras model to be able to add transformation
    steps to convert for example input and output representations to fit MD programs like :obj:`ase` .
    The :obj:`MolDynamicsModelPredictor` receives a :obj:`MemoryGraphList` in call and returns a
    :obj:`MemoryGraphList` .


    """

    def __init__(self,
                 model: ks.models.Model = None,
                 model_inputs: Union[list, dict] = None,
                 model_outputs: Union[list, dict] = None,
                 graph_preprocessors: List[Callable] = None,
                 graph_postprocessors: List[Callable] = None,
                 store_last_input: bool = False,
                 store_last_output: bool = False,
                 copy_graphs_in_store: bool = False,
                 use_predict: bool = False,
                 predict_verbose: Union[str, int] = 0,
                 batch_size: int = 32,
                 update_from_last_input: list = None,
                 update_from_last_input_skip: int = None,
                 ):
        r"""Initialize :obj:`MolDynamicsModelPredictor` class.

        Args:
            model (tf.keras.Model): Single trained keras model.
            model_inputs (list, dict): List or single dictionary for model inputs.
            model_outputs (list, dict): List of model output names or dictionary of output mappings from keras model
                output to the names in the return :obj:`GraphDict` .
            graph_preprocessors (list): List of graph preprocessors, see :obj:`kgcnn.graph.preprocessor` .
            graph_postprocessors (list): List of graph postprocessors, see :obj:`kgcnn.graph.postprocessor` .
            use_predict (bool): Whether to use :obj:`model.predict()` or call method :obj:`model()` .
            batch_size (int): Optional batch size for prediction.
            store_last_input (bool): Whether to store last input graph list for model input. Default is False.
            store_last_output (bool): Whether to store last output graph list from model output. Default is False.
            copy_graphs_in_store (bool): Whether to make a copy of the graph lists or keep reference. Default is False.
            update_from_last_input (list): List of graph properties to copy from last input into current input.
                This is placed before graph preprocessors. Default is None.
            update_from_last_input_skip (int): If set to a value, this will skip the update from last input at
                given number of calls. Uses counter. Default is None.
        """
        if graph_preprocessors is None:
            graph_preprocessors = []
        if graph_postprocessors is None:
            graph_postprocessors = []
        self.model = model
        self.model_inputs = model_inputs
        self.model_outputs = model_outputs
        self.graph_preprocessors = [deserialize(gp) if isinstance(gp, dict) else gp for gp in graph_preprocessors]
        self.graph_postprocessors = [deserialize(gp) if isinstance(gp, dict) else gp for gp in graph_postprocessors]
        self.batch_size = batch_size
        self.use_predict = use_predict
        self.store_last_input = store_last_input
        self.store_last_output = store_last_output
        self.copy_graphs_in_store = copy_graphs_in_store
        self.update_from_last_input = update_from_last_input
        self.update_from_last_input_skip = update_from_last_input_skip
        self.predict_verbose = predict_verbose

        self._last_input = None
        self._last_output = None
        self._counter = 0

    def load(self, file_path: str):
        raise NotImplementedError("Not yet supported.")

    def save(self, file_path: str):
        raise NotImplementedError("Not yet supported.")

    @tf.function
    def _call_model_(self, tensor_input):
        return self.model(tensor_input, training=False)

    @staticmethod
    def _translate_properties(properties, translation) -> dict:
        """Translate general model output.

        Args:
            properties (list, dict): List of properties or dictionary of properties.
            translation (str, list, dict): List of names or dictionary of name mappings like '{new_name: old_name}'.

        Returns:
            dict: Return dictionary with keys from translation.
        """

        if isinstance(translation, list):
            assert isinstance(properties, list), "With '%s' require list for '%s'." % (translation, properties)
            output = {key: properties[i] for i, key in enumerate(translation)}
        elif isinstance(translation, dict):
            assert isinstance(properties, dict), "With '%s' require dict for '%s'." % (translation, properties)
            output = {key: properties[value] for key, value in translation.items()}
        elif isinstance(translation, str):
            assert not isinstance(properties, (list, dict)), "Must be array-like for str '%s'." % properties
            output = {translation: properties}
        else:
            raise TypeError("'%s' output translation must be 'str', 'dict' or 'list'." % properties)
        return output

    def __call__(self, graph_list: MemoryGraphList) -> MemoryGraphList:
        """Prediction of the model wrapper.

        Args:
            graph_list (MemoryGraphList): List of graphs to predict e.g. energies and forces.

        Returns:
            MemoryGraphList: List of general return graph dictionaries from model output.
        """

        num_samples = len(graph_list)

        skip = self._counter % self.update_from_last_input_skip == 0 if self.update_from_last_input_skip else False
        if self.update_from_last_input is not None and self._last_input is not None and not skip:
            for i in range(num_samples):
                for prop in self.update_from_last_input:
                    graph_list[i].set(prop, self._last_input[i].get(prop))

        for gp in self.graph_preprocessors:
            for i in range(num_samples):
                graph_list[i].apply_preprocessor(gp)

        if self.store_last_input:
            if self.copy_graphs_in_store:
                self._last_input = graph_list.copy()
            else:
                self._last_input = graph_list

        tensor_input = graph_list.tensor(self.model_inputs)

        if not self.use_predict:
            tensor_output = self._call_model_(tensor_input)
        else:
            tensor_output = self.model.predict(tensor_input, batch_size=self.batch_size, verbose=self.predict_verbose)

        # Translate output. Mapping of model dict or list to dict for required calculator.
        tensor_dict = self._translate_properties(tensor_output, self.model_outputs)

        # Cast to numpy output and apply postprocessors.
        output_list = []
        for i in range(num_samples):
            temp_dict = {
                key: np.array(value[i]) for key, value in tensor_dict.items()
            }
            temp_dict = GraphDict(temp_dict)
            for mp in self.graph_postprocessors:
                post_temp = mp(graph=temp_dict, pre_graph=graph_list[i])
                temp_dict.update(post_temp)
            output_list.append(temp_dict)

        if self.store_last_output:
            if self.copy_graphs_in_store:
                self._last_output = output_list.copy()
            else:
                self._last_output = output_list

        # Increase counter.
        self._counter += 1

        return MemoryGraphList(output_list)

    def _test_timing(self, graph_list: MemoryGraphList, repetitions: int = 100) -> float:
        """Evaluate timing for prediction.

        Args:
            graph_list (MemoryGraphList): List of graphs to predict e.g. energies and forces.

        Returns:
            float: Time for one call.
        """
        assert repetitions >= 1, "Repetitions must be number of calls."
        start = time.process_time()
        for _ in range(repetitions):
            self.__call__(graph_list)
        stop = time.process_time()
        return float(stop-start)/repetitions



import tensorflow as tf
import numpy as np
import math
from kgcnn.layers.base import GraphBaseLayer
from kgcnn.layers.gather import GatherNodesSelection
# from kgcnn.layers.gather import GatherNodesOutgoing, GatherNodesIngoing
from kgcnn.layers.geom import NodeDistanceEuclidean, NodePosition
from kgcnn.layers.pooling import RelationalPoolingLocalEdges
# from kgcnn.layers.pooling import PoolingLocalEdges
from kgcnn.layers.modules import LazyMultiply, LazySubtract
# from kgcnn.layers.modules import ExpandDims

ks = tf.keras


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ACSFG2')
class ACSFG2(GraphBaseLayer):
    r"""Atom-centered symmetry functions (ACSF) for high-dimensional neural network potentials (HDNNPs).

    `Jörg Behler, The Journal of Chemical Physics 134, 074106 (2011)
    <https://aip.scitation.org/doi/full/10.1063/1.3553717>`_

    This layer implements the radial part :math:`G_{i}^{2}` :

    .. math::

        G_{i}^{2} = \sum_{j \neq i} \; e^{−\eta \, (r_{ij} − \mu)^{2} } \; f_{ij}

    Here, for each atom type there is a set of parameters :math:`\eta` and :math:`\mu` and cutoff.
    The cutoff function :math:`f_ij = f_c(r_{ij})` is given by:

    .. math::

        f_c(r_{ij}) = 0.5 [\cos{\frac{\pi r_{ij}}{R_c}} + 1]

    In principle these parameters can be made trainable. The above sum is conducted for each atom type.

    Example:

    .. code-block:: python

        import tensorflow as tf
        from kgcnn.layers.conv.acsf_conv import ACSFG2
        layer = ACSFG2(
            eta_rs_rc=[[[0.0, 0.0, 8.0], [1.0, 0.0, 8.0]],[[0.0, 0.0, 8.0], [1.0, 0.0, 8.0]]],
            element_mapping=[1, 6]
        )
        z = tf.ragged.constant([[1, 6]], ragged_rank=1)
        xyz = tf.ragged.constant([[[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]]], ragged_rank=1, inner_shape=(3,))
        eij = tf.ragged.constant([[[0,1], [1, 0]]], ragged_rank=1, inner_shape=(2,))
        rep_i = layer([z, xyz, eij])

    """

    _max_atomic_number = 96

    def __init__(self,
                 eta_rs_rc: list,
                 element_mapping: list,
                 add_eps: bool = False,
                 param_constraint=None, param_regularizer=None, param_initializer="zeros",
                 param_trainable: bool = False,
                 **kwargs):
        r"""Initialize layer.

        .. note::

            You can use simpler :obj:`make_param_table` method to generate `eta_zeta_lambda_rc` argument.

        Args:
            eta_rs_rc (list, np.ndarray): List of shape `(N, N, m, 3)` or `(N, m, 3)` where `N` are the considered
                atom types and m the number of representations. Tensor output will be shape `(batch, None, N*m)` .
                In the last dimension are the values for :math:`eta`, :math:`R_s` and :math:`R_c` .
            element_mapping (list): Atomic numbers of elements in :obj:`eta_rs_rc` , must have shape `(N, )` .
                Should not contain duplicate elements.
            add_eps (bool): Whether to add epsilon. Default is False.
            param_constraint: Parameter constraint for weights. Default is None.
            param_regularizer: Parameter regularizer for weights. Default is None.
            param_initializer: Parameter initializer for weights. Default is "zeros".
            param_trainable (bool): Parameter make trainable. Default is False.
        """
        super(ACSFG2, self).__init__(**kwargs)
        # eta_rs_rc of shape (N, N, m, 3) with m combinations of eta, rs, rc
        # or simpler (N, m, 3) where we repeat an additional N dimension assuming same parameter of source.
        self.eta_rs_rc = np.array(eta_rs_rc)
        assert len(self.eta_rs_rc.shape) in [3, 4], "Require `eta_rs_rc` of shape `(N, N, m, 3)` or `(N, m, 3)`"
        self.use_target_set = (len(self.eta_rs_rc.shape) == 4)
        self.num_relations = self.eta_rs_rc.shape[1] if self.use_target_set else self.eta_rs_rc.shape[0]
        self.element_mapping = np.array(element_mapping, dtype="int")  # of shape (N, ) with atomic number for eta_rs_rc
        self.reverse_mapping = np.empty(self._max_atomic_number, dtype="int")
        self.reverse_mapping.fill(np.iinfo(self.reverse_mapping.dtype).max)
        for i, pos in enumerate(self.element_mapping):
            self.reverse_mapping[pos] = i
        self.add_eps = add_eps

        self.lazy_mult = LazyMultiply()
        self.layer_pos = NodePosition()
        self.layer_gather = GatherNodesSelection([0, 1])
        self.layer_dist = NodeDistanceEuclidean(add_eps=add_eps)
        self.pool_sum = RelationalPoolingLocalEdges(num_relations=self.num_relations, pooling_method="sum")

        # We can do this in init since weights do not depend on input shape.
        self.param_initializer = ks.initializers.deserialize(param_initializer)
        self.param_regularizer = ks.regularizers.deserialize(param_regularizer)
        self.param_constraint = ks.constraints.deserialize(param_constraint)
        self.param_trainable = param_trainable

        self.weight_eta_rs_rc = self.add_weight(
            "eta_rs_rc",
            shape=self.eta_rs_rc.shape,
            initializer=self.param_initializer,
            regularizer=self.param_regularizer,
            constraint=self.param_constraint,
            dtype=self.dtype, trainable=self.param_trainable
        )
        self.weight_reverse_mapping = self.add_weight(
            "reverse_mapping",
            shape=(self._max_atomic_number,),
            initializer=self.param_initializer,
            regularizer=self.param_regularizer,
            constraint=self.param_constraint,
            dtype="int64", trainable=False
        )

        self.set_weights([self.eta_rs_rc, self.reverse_mapping])

    @staticmethod
    def make_param_table(eta: list, rs: list, rc: float, elements: list, **kwargs):
        r"""Simplified method to generate a parameter table and input for this layer based on a list of values for
        :math:`R_c` and :math:`\eta` etc.

        Args:
            eta (list): List of etas.
            rs (list): List of rs.
            rc (float): Single Cutoff value.
            elements (list): List of elements.

        Returns:
            dict: Kwargs input for this layer.
        """
        eta_rs_rc = [(et, Rs, rc) for Rs in rs for et in eta]
        elements = np.sort(elements)
        params = np.broadcast_to(eta_rs_rc, (len(elements), len(eta_rs_rc), 3))
        return {"eta_rs_rc": params, "element_mapping": elements, **kwargs}

    def _find_atomic_number_maps(self, inputs):
        return tf.gather(self.weight_reverse_mapping, inputs, axis=0)

    def _find_params_per_bond(self, inputs: list):
        zi_map, zj_map = inputs
        if self.use_target_set:
            params = tf.gather(tf.gather(self.weight_eta_rs_rc, zi_map, axis=0), zj_map, axis=1, batch_dims=1)
        else:
            # Atomic specific for j but not i.
            params = tf.gather(self.weight_eta_rs_rc, zj_map, axis=0)
        return params

    @staticmethod
    def _compute_fc(inputs: tf.Tensor):
        rij, params = inputs
        cutoff = tf.gather(params, 2, axis=-1)
        fc = tf.clip_by_value(tf.broadcast_to(rij, tf.shape(cutoff)), -cutoff, cutoff)
        fc = (tf.math.cos(fc * math.pi / cutoff) + 1.0) * 0.5
        # fc = tf.where(tf.abs(inputs) < self.cutoff, fc, tf.zeros_like(fc))
        return fc

    @staticmethod
    def _compute_gaussian_expansion(inputs: tf.Tensor):
        rij, params = inputs
        eta, mu = tf.gather(params, 0, axis=-1), tf.gather(params, 1, axis=-1)
        arg = tf.square(rij - mu) * eta
        return tf.exp(-arg)

    @staticmethod
    def _flatten_relations(inputs):
        input_shape = tf.shape(inputs)
        flatten_shape = tf.concat(
            [input_shape[:1], tf.constant([inputs.shape[1] * inputs.shape[2]], dtype=input_shape.dtype)], axis=0)
        return tf.reshape(inputs, flatten_shape)

    def build(self, input_shape):
        super(ACSFG2, self).build(input_shape)

    def call(self, inputs, mask=None, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [z, xyz, ij]

                - z (tf.RaggedTensor): Atomic numbers of shape (batch, [N])
                - xyz (tf.RaggedTensor): Node coordinates of shape (batch, [N], 3)
                - ij (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)

            mask: Boolean mask for inputs. Not used. Defaults to None.

        Returns:
            tf.RaggedTensor: Atomic representation of shape `(batch, None, units)` .
        """
        z, xyz, eij = self.assert_ragged_input_rank(inputs, mask=mask, ragged_rank=1)
        z = self.map_values(tf.cast, z, dtype=eij.dtype)
        xi, xj = self.layer_pos([xyz, eij], **kwargs)
        rij = self.layer_dist([xi, xj], **kwargs)
        zi, zj = self.layer_gather([z, eij])
        zi_map = self.map_values(self._find_atomic_number_maps, zi)
        zj_map = self.map_values(self._find_atomic_number_maps, zj)
        params_per_bond = self.map_values(self._find_params_per_bond, [zi_map, zj_map])
        fc = self.map_values(self._compute_fc, [rij, params_per_bond])
        gij = self.map_values(self._compute_gaussian_expansion, [rij, params_per_bond])
        rep = self.lazy_mult([gij, fc], **kwargs)
        pooled = self.pool_sum([xyz, rep, eij, zj_map], **kwargs)
        return self.map_values(self._flatten_relations, pooled)

    def get_config(self):
        config = super(ACSFG2, self).get_config()
        config.update({
            "eta_rs_rc": self.eta_rs_rc.tolist(),
            "element_mapping": self.element_mapping.tolist(),
            "add_eps": self.add_eps,
            "param_constraint": ks.constraints.serialize(self.param_constraint),
            "param_regularizer": ks.regularizers.serialize(self.param_regularizer),
            "param_initializer": ks.initializers.serialize(self.param_initializer),
            "param_trainable": self.param_trainable
        })
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ACSFG4')
class ACSFG4(GraphBaseLayer):
    r"""Atom-centered symmetry functions (ACSF) for high-dimensional neural network potentials (HDNNPs).

    `Jörg Behler, The Journal of Chemical Physics 134, 074106 (2011)
    <https://aip.scitation.org/doi/full/10.1063/1.3553717>`_

    This layer implements the angular part :math:`G_{i}^{4}` :

    .. math::

        :math:`G_{i}^{4}` =  \; \sum_{j\neq i} \sum_{k\neq j,i} \; 2^{1−\zeta}
        (1 + \lambda\, \cos{\theta_{ijk}})^\zeta\;
        \times \; e^{−\eta r_{ij}^{2}} \; e^{−\eta r_{ik}^{2}} \; e^{−\eta r_{jk}^{2}} \;
        \times \; f_{ij} \; f_{ik} \; f_{jk}

    Here, for each atom type there is a set of parameters :math:`\eta` , :math:`\mu` , :math:`\lambda`
    and :math:`\zeta`.
    The cutoff function :math:`f_ij = f_c(r_{ij})` is given by:

    Example:

    .. code-block:: python

        import tensorflow as tf
        from kgcnn.layers.conv.acsf_conv import ACSFG4
        layer = ACSFG4(
            eta_zeta_lambda_rc=[[[0.0, 1.0, -1.0, 8.0]],[[0.0, 1.0, -1.0, 8.0]]],
            element_mapping=[1, 6],
            keep_pair_order=False
        )
        z = tf.ragged.constant([[1, 6, 6]], ragged_rank=1)
        xyz = tf.ragged.constant([[[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]]], ragged_rank=1, inner_shape=(3,))
        ijk = tf.ragged.constant([[[0, 1, 2], [1, 0, 2], [2, 0, 1]]], ragged_rank=1, inner_shape=(3,))
        rep_i = layer([z, xyz, ijk])

    """

    _max_atomic_number = 96

    def __init__(self, eta_zeta_lambda_rc: list,
                 element_mapping: list,
                 element_pair_mapping: list = None,
                 add_eps: bool = False,
                 keep_pair_order: bool = False,
                 multiplicity: float = None,
                 param_initializer="zeros", param_regularizer=None, param_constraint=None,
                 param_trainable: bool = False,
                 **kwargs):
        r"""Initialize layer.

        .. note::

            You can use simpler :obj:`make_param_table` method to generate `eta_zeta_lambda_rc` argument.

        Args:
            eta_zeta_lambda_rc: A list of parameters of shape `(N, M, m,4)` or simply `(M, m, 4)` where `m`
                represents the number of parameter sets, `N` the number of different atom types (set with
                :obj:`element_mapping` ) but which is optional, then all elements share parameters. And `M`
                being the number of angle combinations that can occur. By default, if order is ignored, this
                will be :math:`M=N(N+1)/2` combinations.
            element_mapping (list): Atomic numbers of elements in :obj:`eta_zeta_lambda_rc` ,
                must have shape `(N, )` . Should not contain duplicate elements.
            element_pair_mapping: Atomic pairs for :obj:`eta_zeta_lambda_rc` , where each entry contains atomic
                numbers. Must have shape `(M, 2)` . Default this is generated from N*(N+1)/2 combinations or
                N*N combinations if :obj:`keep_pair_order` is `False`. Can be set manually but must match shape.
            keep_pair_order (bool): Whether to have parameters for order atom pairs that make an angle.
                Default is False.
            multiplicity (float): Angle term is divided by multiplicity, if not None. Default is None.
            add_eps (bool): Whether to add epsilon. Default is False.
            param_constraint: Parameter constraint for weights. Default is None.
            param_regularizer: Parameter regularizer for weights. Default is None.
            param_initializer: Parameter initializer for weights. Default is "zeros".
            param_trainable (bool): Parameter make trainable. Default is False.
        """
        super(ACSFG4, self).__init__(**kwargs)
        self.add_eps = add_eps
        self.multiplicity = multiplicity
        self.keep_pair_order = keep_pair_order
        self.eta_zeta_lambda_rc = np.array(eta_zeta_lambda_rc, dtype="float")
        assert len(self.eta_zeta_lambda_rc.shape) in [3, 4], "Require `eta_zeta_lambda_rc` rank 3 or 4."
        self.use_target_set = (len(self.eta_zeta_lambda_rc.shape) == 4)
        if self.use_target_set:
            self.num_relations = self.eta_zeta_lambda_rc.shape[1]
        else:
            self.num_relations = self.eta_zeta_lambda_rc.shape[0]
        self.element_mapping = np.array(element_mapping, dtype="int")  # of shape (N, ) with atomic number.
        if element_pair_mapping is None:
            element_pair_index = np.expand_dims(self.element_mapping, axis=-1)
            self.element_pair_mapping = np.concatenate([
                    np.repeat(np.expand_dims(element_pair_index, axis=0), len(self.element_mapping), axis=0),
                    np.repeat(np.expand_dims(element_pair_index, axis=1), len(self.element_mapping), axis=1)
                ], axis=-1
            ).reshape((-1, 2))
            if not self.keep_pair_order:
                self.element_pair_mapping = np.sort(self.element_pair_mapping, axis=-1)
                self.element_pair_mapping = self.element_pair_mapping[
                    np.sort(np.unique(self.element_pair_mapping, axis=0, return_index=True)[1])]
        else:
            self.element_pair_mapping = np.array(element_pair_mapping, dtype="int")
        assert len(self.element_pair_mapping.shape) == 2 and self.element_pair_mapping.shape[1] == 2
        assert self.element_pair_mapping.shape[0] == self.num_relations

        self.reverse_mapping = np.empty(self._max_atomic_number, dtype="int")
        self.reverse_mapping.fill(np.iinfo(self.reverse_mapping.dtype).max)
        for i, pos in enumerate(self.element_mapping):
            self.reverse_mapping[pos] = i

        self.reverse_pair_mapping = np.empty((self._max_atomic_number, self._max_atomic_number), dtype="int")
        self.reverse_pair_mapping.fill(np.iinfo(self.reverse_pair_mapping.dtype).max)
        for i, pos in enumerate(self.element_pair_mapping):
            self.reverse_pair_mapping[pos[0], pos[1]] = i
            if not self.keep_pair_order:
                self.reverse_pair_mapping[pos[1], pos[0]] = i

        # Sub-layer.
        self.lazy_mult = LazyMultiply()
        self.layer_pos = NodePosition(selection_index=[0, 1, 2])
        self.layer_dist = NodeDistanceEuclidean(add_eps=add_eps)
        self.pool_sum = RelationalPoolingLocalEdges(num_relations=self.num_relations, pooling_method="sum")
        self.lazy_sub = LazySubtract()
        self.layer_gather = GatherNodesSelection(selection_index=[0, 1, 2])

        # We can do this in init since weights do not depend on input shape.
        self.param_initializer = ks.initializers.deserialize(param_initializer)
        self.param_regularizer = ks.regularizers.deserialize(param_regularizer)
        self.param_constraint = ks.constraints.deserialize(param_constraint)
        self.param_trainable = param_trainable

        self.weight_eta_zeta_lambda_rc = self.add_weight(
            "eta_zeta_lambda_rc",
            shape=self.eta_zeta_lambda_rc.shape,
            initializer=self.param_initializer,
            regularizer=self.param_regularizer,
            constraint=self.param_constraint,
            dtype=self.dtype, trainable=self.param_trainable
        )
        self.weight_reverse_mapping = self.add_weight(
            "reverse_mapping",
            shape=(self._max_atomic_number,),
            initializer=self.param_initializer,
            regularizer=self.param_regularizer,
            constraint=self.param_constraint,
            dtype="int64", trainable=False
        )
        self.weight_reverse_pair_mapping = self.add_weight(
            "reverse_pair_mapping",
            shape=(self._max_atomic_number, self._max_atomic_number),
            initializer=self.param_initializer,
            regularizer=self.param_regularizer,
            constraint=self.param_constraint,
            dtype="int64", trainable=False
        )
        self.set_weights([self.eta_zeta_lambda_rc, self.reverse_mapping, self.reverse_pair_mapping])

    @staticmethod
    def make_param_table(eta: list, zeta: list, lamda: list, rc: float, elements: list, **kwargs):
        r"""Simplified method to generate a parameter table and input for this layer based on a list of values for
        :math:`R_c` and :math:`\eta` etc.

        Args:
            eta (list): List of etas.
            zeta (list): List of zeta.
            lamda (list): List of lamda.
            rc (float): Single Cutoff value.
            elements (list): List of elements.

        Returns:
            dict: Kwargs input for this layer.
        """
        eta_zeta_lambda_rc = [[eta, z, la, rc] for eta in eta for z in zeta for la in lamda]
        elements = np.sort(elements)
        params = np.broadcast_to(
            eta_zeta_lambda_rc, (int(len(elements) * (len(elements) + 1) / 2), len(eta_zeta_lambda_rc), 4))
        return {"eta_zeta_lambda_rc": params, "element_mapping": elements, "element_pair_mapping": None, **kwargs}

    def _find_atomic_number_maps(self, inputs):
        return tf.gather(self.weight_reverse_mapping, inputs, axis=0)

    def _find_atomic_number_pair_maps(self, inputs):
        zj, zk = inputs
        return tf.gather(tf.gather(self.weight_reverse_pair_mapping, zj, axis=0), zk, axis=1, batch_dims=1)

    def _find_params_per_bond(self, inputs: list):
        zi_map, zjk_map = inputs
        if self.use_target_set:
            params = tf.gather(tf.gather(self.weight_eta_zeta_lambda_rc, zi_map, axis=0), zjk_map, axis=1, batch_dims=1)
        else:
            # Atomic specific for j,k but not i.
            params = tf.gather(self.weight_eta_zeta_lambda_rc, zjk_map, axis=0)
        return params

    @staticmethod
    def _compute_fc(inputs: tf.Tensor):
        rij, params = inputs
        cutoff = tf.gather(params, 3, axis=-1)
        fc = tf.clip_by_value(tf.broadcast_to(rij, tf.shape(cutoff)), -cutoff, cutoff)
        fc = (tf.math.cos(fc * np.pi / cutoff) + 1.0) * 0.5
        # fc = tf.where(tf.abs(inputs) < self.cutoff, fc, tf.zeros_like(fc))
        return fc

    @staticmethod
    def _compute_gaussian_expansion(inputs: tf.Tensor):
        rij, params = inputs
        eta = tf.gather(params, 0, axis=-1)
        # mu = tf.gather(params, 1, axis=-1)
        arg = tf.square(rij) * eta
        return tf.exp(-arg)

    def _compute_pow_cos_angle_(self, inputs: list):
        vij, vik, rij, rik, params = inputs
        lamda, zeta = tf.gather(params, 2, axis=-1), tf.gather(params, 1, axis=-1)
        cos_theta = tf.reduce_sum(vij * vik, axis=-1, keepdims=True) / rij / rik
        cos_term = cos_theta * lamda + 1.0
        cos_term = tf.pow(cos_term, zeta)
        scale = tf.ones_like(cos_term) * 2.0
        scaled_cos_term = tf.pow(scale, 1.0 - zeta) * cos_term
        if self.multiplicity is not None:
            scaled_cos_term = scaled_cos_term/self.multiplicity
        return scaled_cos_term

    @staticmethod
    def _flatten_relations(inputs):
        input_shape = tf.shape(inputs)
        flatten_shape = tf.concat(
            [input_shape[:1], tf.constant([inputs.shape[1] * inputs.shape[2]], dtype=input_shape.dtype)], axis=0)
        return tf.reshape(inputs, flatten_shape)

    def build(self, input_shape):
        super(ACSFG4, self).build(input_shape)

    def call(self, inputs, mask=None, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [z, xyz, ijk]

                - z (tf.RaggedTensor): Atomic numbers of shape (batch, [N])
                - xyz (tf.RaggedTensor): Node coordinates of shape (batch, [N], 3)
                - ijk (tf.RaggedTensor): Angle indices referring to nodes of shape (batch, [M], 3)

            mask: Boolean mask for inputs. Not used. Defaults to None.

        Returns:
            tf.RaggedTensor: Atomic representation of shape `(batch, None, units)` .
        """
        z, xyz, ijk = self.assert_ragged_input_rank(inputs, mask=mask, ragged_rank=1)
        z = self.map_values(tf.cast, z, dtype=ijk.dtype)
        zi, zj, zk = self.layer_gather([z, ijk], **kwargs)
        xi, xj, xk = self.layer_pos([xyz, ijk], **kwargs)
        zi_map = self.map_values(self._find_atomic_number_maps, zi)
        zjk_map = self.map_values(self._find_atomic_number_pair_maps, [zj, zk])
        params_per_bond = self.map_values(self._find_params_per_bond, [zi_map, zjk_map])
        rij = self.layer_dist([xi, xj], **kwargs)
        rik = self.layer_dist([xi, xk], **kwargs)
        rjk = self.layer_dist([xj, xk], **kwargs)
        fij = self.map_values(self._compute_fc, [rij, params_per_bond])
        fik = self.map_values(self._compute_fc, [rik, params_per_bond])
        fjk = self.map_values(self._compute_fc, [rjk, params_per_bond])
        gij = self.map_values(self._compute_gaussian_expansion, [rij, params_per_bond])
        gik = self.map_values(self._compute_gaussian_expansion, [rik, params_per_bond])
        gjk = self.map_values(self._compute_gaussian_expansion, [rjk, params_per_bond])
        vij = self.lazy_sub([xi, xj], **kwargs)
        vik = self.lazy_sub([xi, xk], **kwargs)
        pow_cos_theta = self.map_values(self._compute_pow_cos_angle_, [vij, vik, rij, rik, params_per_bond])
        rep = self.lazy_mult([pow_cos_theta, gij, gik, gjk, fij, fik, fjk], **kwargs)
        pool_ang = self.pool_sum([xyz, rep, ijk, zjk_map], **kwargs)
        return self.map_values(self._flatten_relations, pool_ang)

    def get_config(self):
        config = super(ACSFG4, self).get_config()
        config.update({
            "eta_zeta_lambda_rc": self.eta_zeta_lambda_rc,
            "add_eps": self.add_eps,
            "element_mapping": self.element_mapping,
            "keep_pair_order": self.keep_pair_order,
            "multiplicity": self.multiplicity,
            "element_pair_mapping": self.element_pair_mapping,
            "param_trainable": self.param_trainable,
            "param_constraint": ks.constraints.serialize(self.param_constraint),
            "param_regularizer": ks.regularizers.serialize(self.param_regularizer),
            "param_initializer": ks.initializers.serialize(self.param_initializer)
        })
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ACSFConstNormalization')
class ACSFConstNormalization(GraphBaseLayer):
    """Simple layer to add a constant feature normalization to conform with reference code."""

    def __init__(self, std=1.0, mean=0.0, **kwargs):
        super(ACSFConstNormalization, self).__init__(**kwargs)
        self._np_std = np.array(std)
        self._np_mean = np.array(mean)
        # Could do some shape checks of std and mean here.
        self._tf_std = tf.constant(self._np_std.tolist())
        self._tf_mean = tf.constant(self._np_mean.tolist())

    def _scale_representation(self, inputs):
        return (inputs-tf.cast(self._tf_mean, inputs.dtype))/tf.cast(self._tf_std, inputs.dtype)

    def call(self, inputs, mask=None, **kwargs):
        r"""Forward pass.

        Args:
            inputs: Tensor of ACSF representation of shape `(batch, [None], units)` .
            mask: Boolean mask for inputs. Not used. Defaults to None.

        Returns:
            tf.RaggedTensor: Normalized atomic representation of shape `(batch, [None], units)` .
        """
        return self.map_values(self._scale_representation, inputs)

    def get_config(self):
        config = super(ACSFConstNormalization, self).get_config()
        config.update({
            "mean": self._np_mean.tolist(),
            "std": self._np_std.tolist()
        })
        return config
